from collections import OrderedDict
from marshmallow.schema import Schema
from pydantic import BaseModel
import logging
import os
import re
from typing import get_type_hints
from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from enum import Enum

import goblet

from goblet.handlers.handler import Handler
from goblet.handlers.plugins.pydantic import PydanticPlugin
from goblet.utils import get_g_dir
from goblet.common_cloud_actions import deploy_apigateway, destroy_apigateway
from goblet.permissions import gcp_generic_resource_permissions
from goblet.errors import GobletRouteNotFoundError

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class Routes(Handler):
    """Either cloudrun routes type or Api Gateway instance, which includes api, api config, api gateway instances
    https://cloud.google.com/api-gateway
    """

    resource_type = "routes"
    valid_backends = ["cloudfunction", "cloudrun", "cloudfunctionv2"]
    required_apis = ["cloudfunctions", "apigateway"]
    permissions = [
        "apigateway.operations.get",
        *gcp_generic_resource_permissions("apigateway", "apiconfigs"),
        *gcp_generic_resource_permissions("apigateway", "apis"),
        *gcp_generic_resource_permissions("apigateway", "gateways"),
    ]

    def __init__(
        self,
        name,
        backend,
        versioned_clients=None,
        cors=None,
        resources=None,
        routes_type="apigateway",
    ):
        super(Routes, self).__init__(
            name=name,
            versioned_clients=versioned_clients,
            resources=resources,
            backend=backend,
        )
        self.name = self.format_name(name)
        self.cors = cors or {}
        self.routes_type = routes_type
        self.marshmallow_attribute_function = None

    def format_name(self, name):
        # ([a-z0-9-.]+) for api gateway name
        return name.replace("_", "-")

    def register(self, name, func, kwargs):
        path = kwargs.pop("path")
        methods = kwargs.pop("methods")
        kwargs = kwargs.pop("kwargs")
        if not kwargs.get("cors"):
            kwargs["cors"] = self.cors
        path_entries = self.resources.get(path, {})
        if kwargs["cors"] and "OPTIONS" in methods:
            raise ValueError(
                "Route entry cannot have both cors=True and methods=['OPTIONS', ...] configured."
            )
        for method in methods:
            if path_entries.get(method):
                raise ValueError(
                    "Duplicate method: '%s' detected for route: '%s'\n"
                    'between view functions: "%s" and "%s". A specific '
                    "method may only be specified once for "
                    "a particular path."
                    % (method, path, self.resources[path][method].function_name, name)
                )
            entry = RouteEntry(func, name, path, method, **kwargs)
            path_entries[method] = entry
        # Add OPTIONS if cors set
        if kwargs["cors"]:
            entry = RouteEntry(handle_cors_options, name, path, "OPTIONS", **kwargs)
            path_entries["OPTIONS"] = entry
        self.resources[path] = path_entries

    def __call__(self, request, context=None):
        method = request.method
        path = request.path
        entry = self.resources.get(path, {}).get(method)
        if not entry:
            # search param paths "/{PARAM}"
            for p in self.resources:
                if "{" in p and self._matched_path(p, path):
                    entry = self.resources.get(p, {}).get(method)
        if not entry:
            raise GobletRouteNotFoundError(f"No route found for {path} with {method}")
        return entry(request)

    @staticmethod
    def _matched_path(org_path, path):
        split_org = re.sub(r"{[\w]+}", "", org_path).split("/")
        split_path = path.split("/")
        if len(split_path) != len(split_org):
            return False
        for i, item in enumerate(split_org):
            if item and split_path[i] != item:
                return False
        return True

    def _deploy(self, source=None, entrypoint=None):
        if (
            self.routes_type != "apigateway"
            and self.backend.resource_type.startswith("cloudfunction")
            and self.versioned_clients.cloudfunctions == "v1"
        ):
            raise ValueError(
                f"Cloudfunctions v1 backend is not supported for routes_type {self.routes_type}"
            )
        if len(self.resources) == 0 or self.routes_type != "apigateway":
            return
        log.info("deploying api......")
        base_url = self.backend.http_endpoint
        self.generate_openapi_spec(base_url)
        if (
            self.config
            and self.config.apiConfig
            and self.config.apiConfig.get("gatewayServiceAccount")
        ):
            self.service_accounts = [self.config.apiConfig.get("gatewayServiceAccount")]
        deploy_apigateway(
            self.name,
            self.config,
            self.versioned_clients,
            f"{get_g_dir()}/{self.name}_openapi_spec.yml",
        )
        return

    def destroy(self):
        if len(self.resources) == 0 or self.routes_type != "apigateway":
            return
        destroy_apigateway(self.name, self.versioned_clients)

    def generate_openapi_spec(self, cloudfunction):
        deadline = self.get_timeout(self.config)
        spec = OpenApiSpec(
            self.name,
            cloudfunction,
            security_definitions=self.config.securityDefinitions,
            security=self.config.security,
            marshmallow_attribute_function=self.marshmallow_attribute_function,
            deadline=deadline,
        )
        spec.add_apigateway_routes(self.resources)
        with open(f"{get_g_dir()}/{self.name}_openapi_spec.yml", "w") as f:
            spec.write(f)

    def get_timeout(self, config):
        # get api gateway timeout
        deadline = (config.api_gateway or {}).get("deadline")
        if self.backend.resource_type == "cloudfunction" and not deadline:
            deadline = (config.cloudfunction or {}).get("timeout")
        if self.backend.resource_type == "cloudfunctionv2" and not deadline:
            deadline = (
                (config.cloudfunction_v2 or {})
                .get("serviceConfig", {})
                .get("timeoutSeconds")
            )
        if self.backend.resource_type == "cloudrun" and not deadline:
            deadline = (config.cloudrun_revision or {}).get("timeout")
        # default deadline to 15 seconds, which is gcp api gateway default
        return deadline or 15


PRIMITIVE_MAPPINGS = {str: "string", bool: "boolean", int: "integer"}


class OpenApiSpec:
    def __init__(
        self,
        app_name,
        cloudfunction,
        version="1.0.0",
        security_definitions=None,
        security=None,
        marshmallow_attribute_function=None,
        deadline=15,
        existing_spec=None,
    ):
        self.options = OrderedDict()
        self.app_name = app_name
        self.cloudfunction = cloudfunction
        self.version = version
        self.deadline = deadline
        self.options["swagger"] = "2.0"
        if security_definitions:
            security_definitions = {**security_definitions}
            self.options["securityDefinitions"] = security_definitions
            self.options["security"] = security or list(
                map(lambda s: {s: []}, security_definitions)
            )
        self.options["schemes"] = ["https"]
        self.options["produces"] = ["application/json"]
        marshmallow_plugin = MarshmallowPlugin()
        pydantic_plugin = PydanticPlugin()
        # Support existing spec. Needs to be version "2.0"
        if existing_spec:
            if not existing_spec.get("swagger") == "2.0":
                raise ValueError("API Gateway only supports swagger 2.0")
            self.component_spec = APISpec(
                title=self.app_name,
                version=self.version,
                openapi_version="2.0",
                **existing_spec,
            )
        else:
            self.component_spec = APISpec(
                title=self.app_name,
                version=self.version,
                openapi_version="2.0",
                # Pydantic plugin needs to go first
                plugins=[pydantic_plugin, marshmallow_plugin],
                **self.options,
            )
        if marshmallow_attribute_function:
            marshmallow_plugin.converter.add_attribute_function(
                marshmallow_attribute_function
            )

    def add_component(self, component, **kwargs):
        if component.__name__ in self.component_spec.components.schemas:
            return
        self.component_spec.components.schema(component.__name__, **kwargs)
        # self.spec["definitions"] = self.component_spec.to_dict()["definitions"]

    def add_apigateway_routes(self, apigateway):
        for path, methods in apigateway.items():
            for method, entry in methods.items():
                self.add_route(entry)

    def get_param_type(self, type_info, only_primititves=False):
        if not type_info:
            return {"type": "string"}
        if type_info in PRIMITIVE_MAPPINGS.keys():
            param_type = {"type": PRIMITIVE_MAPPINGS[type_info]}
        elif issubclass(type_info, Enum):
            param_type = {"type": "string", "enum": [e.value for e in type_info]}
        elif issubclass(type_info, Schema) and not only_primititves:
            self.add_component(type_info, schema=type_info)
            param_type = {"$ref": f"#/definitions/{type_info.__name__}"}
        elif issubclass(type_info, BaseModel) and not only_primititves:
            self.add_component(type_info, model=type_info)
            param_type = {"$ref": f"#/definitions/{type_info.__name__}"}
        else:
            raise ValueError(
                f"param_type has type {type_info}. \
                It must be of type {PRIMITIVE_MAPPINGS.values} or a dataclass inheriting from Schema"
            )
        return param_type

    def add_route(self, entry):
        # https://apispec.readthedocs.io/en/latest/quickstart.html add path
        method_spec = OrderedDict()
        method_spec["x-google-backend"] = {
            "address": entry.backend or self.cloudfunction,
            "protocol": "h2",
            "path_translation": "APPEND_PATH_TO_ADDRESS",
            "deadline": entry.deadline or self.deadline,
        }
        method_spec["operationId"] = f"{entry.method.lower()}_{entry.function_name}"

        params = []
        type_hints = get_type_hints(entry.route_function)
        for param in entry.view_args:
            type_info = type_hints.get(param, str)
            param_type = self.get_param_type(type_info, only_primititves=True)

            param_entry = {"in": "path", "name": param, "required": True, **param_type}
            params.append(param_entry)

        if entry.request_body:
            params.append(
                {
                    "in": "body",
                    "name": "requestBody",
                    **self._extract_content(entry.request_body),
                }
            )

        if entry.form_data:
            params.append({"in": "formData", "name": "file", "type": "string"})
            entry.content_types = ["multipart/form-data"]

        if entry.query_params:
            method_spec["parameters"] = []
            for query in entry.query_params:
                if not query.get("schema"):
                    params.append({"in": "query", **query})
                else:
                    params.append(query)
        if params:
            method_spec["parameters"] = params

        return_type = type_hints.get("return")
        content = {}
        if return_type:
            content = self._extract_content(return_type)

        if entry.responses:
            method_spec["responses"] = entry.responses
        else:
            method_spec["responses"] = {
                "200": {"description": "A successful response", **content}
            }
        if entry.security:
            method_spec["security"] = entry.security

        if entry.tags:
            method_spec["tags"] = entry.tags

        if entry.content_types:
            method_spec["consumes"] = entry.content_types

        self.component_spec.path(
            entry.uri_pattern,
            operations={
                entry.method.lower(): dict(
                    **method_spec,
                )
            },
        )

    def _extract_content(self, return_type):
        """
        Return openapi spec response content for the given return type
        """
        if return_type in PRIMITIVE_MAPPINGS.keys():
            return {"schema": {"type": PRIMITIVE_MAPPINGS[return_type]}}
        # list
        if "typing.List" in str(return_type):
            type_info = return_type.__args__[0]
            param_type = self.get_param_type(type_info)
            return {"schema": {"type": "array", "items": {**param_type}}}
        if issubclass(return_type, Schema):
            param_type = self.get_param_type(return_type)
            return {"schema": {**param_type}}
        if issubclass(return_type, BaseModel):
            param_type = self.get_param_type(return_type)
            return {"schema": {**param_type}}

    def add_x_google_backend(self):
        """Add x-google-backend section to custom openapi specs"""
        for path in self.component_spec.options.get("paths", {}).values():
            for method_options in path.values():
                method_options["x-google-backend"] = {
                    "address": self.cloudfunction,
                    "deadline": self.deadline,
                    "path_translation": "APPEND_PATH_TO_ADDRESS",
                    "protocol": "h2",
                }

    def write(self, file):
        file.write(self.component_spec.to_yaml())


_PARAMS = re.compile(r"{\w+}")


class RouteEntry:
    def __init__(
        self,
        route_function,
        function_name,
        path,
        method,
        api_key_required=None,
        content_types=None,
        cors=False,
        **kwargs,
    ):
        self.route_function = route_function
        self.function_name = function_name
        self.uri_pattern = path
        self.method = method
        self.api_key_required = api_key_required
        self.request_body = kwargs.get("request_body")
        self.query_params = kwargs.get("query_params")
        self.form_data = kwargs.get("form_data")
        self.responses = kwargs.get("responses")
        self.backend = kwargs.get("backend")
        self.security = kwargs.get("security")
        self.tags = kwargs.get("tags")
        self.deadline = kwargs.get("deadline")
        #: A list of names to extract from path:
        #: e.g, '/foo/{bar}/{baz}/qux -> ['bar', 'baz']
        self.view_args = self._parse_view_args()
        self.content_types = content_types
        # cors is passed as either a boolean or a CORSConfig object. If it is a
        # boolean it needs to be replaced with a real CORSConfig object to
        # pass the typechecker. None in this context will not inject any cors
        # headers, otherwise the CORSConfig object will determine which
        # headers are injected.
        if cors is True:
            if isinstance(cors, CORSConfig):
                cors = cors
            else:
                cors = CORSConfig()
        elif cors is False:
            cors = None
        self.cors = cors
        self.kwargs = {**kwargs}

    def _extract_view_args(self, path):
        components = path.split("/")
        original_components = self.uri_pattern.split("/")
        matches = 0
        args = {}
        for i, component in enumerate(components):
            if component != original_components[i]:
                args[self.view_args[matches]] = component
                matches += 1
        return args

    def __call__(self, request):
        args = self._extract_view_args(request.path)
        resp = self.route_function(**args)
        return self._apply_cors(resp)

    def _parse_view_args(self):
        if "{" not in self.uri_pattern:
            return []
        # The [1:-1] slice is to remove the braces
        # e.g {foobar} -> foobar
        results = [r[1:-1] for r in _PARAMS.findall(self.uri_pattern)]
        return results

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def _apply_cors(self, resp):
        """Apply cors to Response"""
        if not self.cors:
            return resp
        # Apply to Response Obj
        if isinstance(resp, goblet.Response):
            resp.headers.update(self.cors.get_access_control_headers())
        if isinstance(resp, tuple):
            resp[2].update(self.cors.get_access_control_headers())
        if isinstance(resp, str):
            resp = goblet.Response(resp, headers=self.cors.get_access_control_headers())
        return resp


class CORSConfig(object):
    """A cors configuration to attach to a route."""

    _REQUIRED_HEADERS = ["Content-Type", "Authorization"]

    def __init__(
        self,
        allow_origin="*",
        allow_headers=None,
        expose_headers=None,
        max_age=None,
        allow_credentials=None,
    ):
        self.allow_origin = allow_origin

        if allow_headers is None:
            allow_headers = set(self._REQUIRED_HEADERS)
        else:
            allow_headers = set(allow_headers + self._REQUIRED_HEADERS)
        self._allow_headers = allow_headers

        if expose_headers is None:
            expose_headers = []
        self._expose_headers = expose_headers

        self._max_age = max_age
        self._allow_credentials = allow_credentials

    @property
    def allow_headers(self):
        return ",".join(sorted(self._allow_headers))

    def get_access_control_headers(self):
        headers = {
            "Access-Control-Allow-Origin": self.allow_origin,
            "Access-Control-Allow-Headers": self.allow_headers,
        }
        if self._expose_headers:
            headers.update(
                {"Access-Control-Expose-Headers": ",".join(self._expose_headers)}
            )
        if self._max_age is not None:
            headers.update({"Access-Control-Max-Age": str(self._max_age)})
        if self._allow_credentials is True:
            headers.update({"Access-Control-Allow-Credentials": "true"})

        return headers

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.get_access_control_headers() == other.get_access_control_headers()
            )
        return False


def handle_cors_options(**kwargs):
    """Return 200"""
    return "success"



import logging
import os
import inspect

from googleapiclient.errors import HttpError
from goblet.infrastructures.infrastructure import Infrastructure
from goblet_gcp_client.client import (
    get_default_project,
    get_default_location,
)
from goblet.permissions import gcp_generic_resource_permissions
from goblet.client import VersionedClients


log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class BigQuerySparkStoredProcedure(Infrastructure):
    """
    Cloud Big Query Spark Stored procedures.
    https://cloud.google.com/bigquery/docs/spark-procedures

    Limitations
    https://cloud.google.com/bigquery/docs/spark-procedures#limitations

    """

    resource_type = "bqsparkstoredprocedure"
    required_apis = ["bigquery", "bigqueryconnection"]
    permissions = [
        "bigquery.jobs.create",
        "bigquery.connections.delegate",
        "bigquery.connections.use",
        *gcp_generic_resource_permissions("storage", "objects"),
        *gcp_generic_resource_permissions("storage", "buckets"),
        *gcp_generic_resource_permissions("bigquery", "connections"),
        *gcp_generic_resource_permissions("bigquery", "table"),
        *gcp_generic_resource_permissions("bigquery", "routines"),
    ]

    def register(self, name, kwargs):
        """
        Register in handler resources
        :param name: name of resource
        :param kwargs:
        :return:
        """
        config = (
            self.config.bqsparkstoredprocedure.copy()
            if self.config.bqsparkstoredprocedure
            else {}
        )
        # Routine names must contain only letters, numbers, and underscores, and be at most 256 characters long.
        routine_name = config.get("name", name).replace("-", "_")
        dataset_id = config.get("dataset_id", kwargs["dataset_id"])
        runtime_version = config.get("runtime_version", kwargs["runtime_version"])
        # Func cannot be loaded from config file as it is a function
        func = kwargs.get("func")
        spark_file = config.get("spark_file", kwargs["spark_file"])
        container_image = config.get(
            "container_image", kwargs.get("container_image", None)
        )
        additional_python_files = config.get(
            "additional_python_files", kwargs.get("additional_python_files", [])
        )
        additional_files = config.get(
            "additional_files", kwargs.get("additional_files", [])
        )
        properties = config.get("properties", kwargs.get("properties", {}))

        local_code = False
        if func is not None:
            func = self.stringify_func(func)
            local_code = True

        self.connection_location = config.get(
            "location", kwargs.get("location", get_default_location())
        )
        self.resources[routine_name] = {
            "routine_name": routine_name,
            "dataset_id": dataset_id,
            "func": func,
            "location": self.connection_location,
            "runtime_version": runtime_version,
            "spark_file": spark_file,
            "local_code": local_code,
            "container_image": container_image,
            "additional_python_files": additional_python_files,
            "additional_files": additional_files,
            "properties": properties,
        }
        return True

    def _deploy(self):
        if not self.resources:
            return
        log.info("Deploying bigquery remote functions")

        try:
            self.deploy_bigquery_connection(self.name, self.connection_location)
        except HttpError as exception:
            if exception.resp.status == 409:
                log.info("Connection already created bigquery query: for %s", self.name)
            else:
                log.error("Create connection %s", exception.error_details)
                raise exception

        for _, resource in self.resources.items():
            if not resource["local_code"]:
                self.deploy_bucket(self.name)
                resource["spark_file"] = self.upload_file(
                    resource["spark_file"], self.name
                )
                if resource["additional_python_files"]:
                    for i in range(len(resource["additional_python_files"])):
                        resource["additional_python_files"][i] = self.upload_file(
                            resource["additional_python_files"][i], self.name
                        )
                if resource["additional_files"]:
                    for i in range(len(resource["additional_files"])):
                        resource["additional_files"][i] = self.upload_file(
                            resource["additional_files"][i], self.name
                        )

            create_routine_query = self.create_routine_payload(resource)
            routine_name = resource["routine_name"]
            try:
                self.versioned_clients.bigquery_routines.execute(
                    "insert",
                    params={
                        "body": create_routine_query,
                        "projectId": get_default_project(),
                        "datasetId": resource["dataset_id"],
                    },
                    parent_key="projectId",
                )
                log.info("Created bq routine %s", routine_name)
            except HttpError as exception:
                if exception.resp.status == 409:
                    self.versioned_clients.bigquery_routines.execute(
                        "update",
                        params={
                            "body": create_routine_query,
                            "projectId": get_default_project(),
                            "datasetId": resource["dataset_id"],
                            "routineId": routine_name,
                        },
                        parent_key="projectId",
                    )
                    log.info("Updated Spark Stored Procedure %s", routine_name)
                else:
                    log.error(
                        "Bigquery Spark Stored Procedure couldn't be created "
                        "nor updated name %s with error: %s",
                        routine_name,
                        exception.error_details,
                    )
                    raise exception

    def destroy(self):
        """
        Destroy connection then destroy one by one every routine
        :return:
        """
        if not self.resources:
            return
        self.destroy_bigquery_connection()
        for _, resource in self.resources.items():
            self.destroy_routine(resource["dataset_id"], resource["routine_name"])
            if not resource["local_code"]:
                self.destroy_bucket(self.name)

    def deploy_bigquery_connection(self, connection_name, location):
        """
            Creates (or get if exists) a connection resource with Handler.name
        :param connection_name: name for the connection
        :return:
        """
        connection_id = f"{self.name}"
        resource_type = {"spark": {}}
        try:
            bq_connection = self.versioned_clients.bigquery_connections.execute(
                "create",
                params={"body": resource_type, "connectionId": connection_id},
                parent_schema=f"projects/{get_default_project()}/locations/{location}",
            )
            log.info(f"Created bigquery connection name: {connection_id}")

        except HttpError as exception:
            if exception.resp.status == 409:
                log.info(
                    f"Bigquery connection already exist with name: {connection_name} for {self.name} and location {location}"
                )
                client = self.versioned_clients.bigquery_connections
                bq_connection = client.execute(
                    "get",
                    params={"name": client.parent + "/connections/" + connection_id},
                    parent=False,
                )
                log.info(f"Returning connection {bq_connection['name']}")
            else:
                log.error(exception.error_details)
                raise exception
        return bq_connection

    def destroy_bigquery_connection(self):
        """
        Destroy bigquery connection, if already exist do nothing
        :return:
        """
        client = self.versioned_clients.bigquery_connections
        try:
            client.execute(
                "delete",
                params={
                    "name": f"projects/{get_default_project()}/locations/{self.connection_location}"
                    + "/connections/"
                    + self.name
                },
                parent=False,
            )
        except HttpError as exception:
            if exception.resp.status == 404:
                log.info(f"Connection {self.name} already destroyed")
            else:
                raise exception
        return True

    def destroy_routine(self, dataset_id, routine_id):
        """
        :param dataset_id:
        :param routine_id:
        """
        try:
            self.versioned_clients.bigquery_routines.execute(
                "delete",
                params={
                    "projectId": get_default_project(),
                    "datasetId": dataset_id,
                    "routineId": routine_id,
                },
                parent=False,
            )
            log.info(f"Destroyed routine {routine_id} for dataset {dataset_id}")
        except HttpError as exception:
            if exception.resp.status == 409:
                log.info(f"Routine {routine_id} already destroyed")
            elif exception.resp.status == 404:
                log.info(f"Routine {routine_id} doesn't exist. already destroyed?")
            else:
                log.error(
                    f"Couldn't destroy {routine_id} for dataset {dataset_id}. {exception.error_details}"
                )
                raise exception

    def create_routine_payload(self, resource):
        """
        Create a routine object according to BigQuery specification
        :param resource: a resource saved in resources in Handler
        :return: a dict representing a routine according to
                https://cloud.google.com/bigquery/docs/reference/rest/v2/routines
        """
        spark_options = {
            "connection": f"projects/{get_default_project()}/locations/{resource['location']}/connections/{self.name}",
            "runtimeVersion": resource["runtime_version"],
            "containerImage": resource["container_image"],
            "properties": resource["properties"],
        }
        routine_reference = {
            "projectId": get_default_project(),
            "datasetId": resource["dataset_id"],
            "routineId": resource["routine_name"],
        }
        language = "PYTHON"

        query_request = {
            "language": language,
            "routineReference": routine_reference,
            "routineType": "PROCEDURE",
            "sparkOptions": spark_options,
        }

        if resource["local_code"]:
            query_request["definitionBody"] = resource["func"]
        else:
            spark_options["mainFileUri"] = resource["spark_file"]
            query_request["sparkOptions"] = spark_options

        if resource["additional_python_files"]:
            spark_options["pyFileUris"] = resource["additional_python_files"]
            query_request["sparkOptions"] = spark_options

        if resource["additional_files"]:
            spark_options["archiveUris"] = resource["additional_files"]
            query_request["sparkOptions"] = spark_options

        log.debug("Routine payload %s", query_request)

        return query_request

    def deploy_bucket(self, bucket_name):
        try:
            log.info(f"creating storage bucket {bucket_name}")
            VersionedClients().storage_buckets.execute(
                "insert",
                params={
                    "project": get_default_project(),
                    "body": {
                        "name": bucket_name,
                    },
                },
            )
            log.info(f"bucket {bucket_name} created")
        except HttpError as e:
            if e.resp.status == 409:
                log.info(f"storage bucket {bucket_name} already exists")

    def upload_file(self, file, bucket_name):
        log.debug(f"gs://{bucket_name}/{file}")
        VersionedClients().storage_objects.execute(
            "insert",
            params={
                "bucket": bucket_name,
                "uploadType": "media",
                "media_body": file,
                "body": {
                    "name": file,
                },
            },
        )
        log.info(f"uploaded file {file} to bucket {bucket_name}")
        return f"gs://{bucket_name}/{file}"

    def destroy_bucket(self, bucket_name):
        # Empty bucket
        log.info(f"emptying storage bucket {bucket_name}")
        objects = VersionedClients().storage_objects.execute(
            "list",
            params={
                "bucket": bucket_name,
            },
        )["items"]
        for obj in objects:
            VersionedClients().storage_objects.execute(
                "delete",
                params={
                    "bucket": bucket_name,
                    "object": obj["name"],
                },
            )
        try:
            log.info(f"deleting storage bucket {bucket_name}")
            VersionedClients().storage_buckets.execute(
                "delete",
                params={
                    "bucket": bucket_name,
                },
            )
            log.info(f"bucket {bucket_name} deleted")
        except HttpError as e:
            if e.resp.status == 404:
                log.info(f"storage bucket {bucket_name} already deleted")

    @staticmethod
    def stringify_func(func):
        lines, _ = inspect.getsourcelines(func)
        if lines[0].startswith("def"):
            lines.pop(0)
        return "".join(map(str.lstrip, lines))



from goblet.infrastructures.infrastructure import Infrastructure
from goblet.permissions import gcp_generic_resource_permissions
from googleapiclient.errors import HttpError
import logging
import os

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class Redis(Infrastructure):
    resource_type = "redis"
    update_keys = ["displayName", "labels", "memorySizeGb", "replicaCount"]
    required_apis = ["redis"]
    permissions = [
        "redis.operations.get",
        gcp_generic_resource_permissions("redis", "instances"),
    ]

    def register(self, name, kwargs):
        self.resources = {"name": name}

    def _deploy(self):
        if not self.resources:
            return
        redis_config = self.config.redis or {}
        req_body = {
            "tier": redis_config.get("tier", "BASIC"),
            "memorySizeGb": redis_config.get("memorySizeGb", 1),
            "labels": self.config.labels,
            **redis_config,
        }
        try:
            resp = self.versioned_clients.redis.execute(
                "create",
                params={"instanceId": self.resources["name"], "body": req_body},
            )
            self.versioned_clients.redis.wait_for_operation(resp["name"])
        except HttpError as e:
            if e.resp.status == 409:
                log.info(f"updating redis {self.resources['name']}")
                if req_body.get("tier") == "BASIC":
                    self.update_keys.remove("replicaCount")
                resp = self.versioned_clients.redis.execute(
                    "patch",
                    parent_key="name",
                    parent_schema="projects/{project_id}/locations/{location_id}/instances/"
                    + self.resources["name"],
                    params={"updateMask": ",".join(self.update_keys), "body": req_body},
                )
                self.versioned_clients.redis.wait_for_operation(resp["name"])
            else:
                raise e

    def destroy(self):
        try:
            if not self.resources:
                return
            resp = self.versioned_clients.redis.execute(
                "delete",
                parent_key="name",
                parent_schema="projects/{project_id}/locations/{location_id}/instances/"
                + self.resources["name"],
            )
            self.versioned_clients.redis.wait_for_operation(resp["name"])
            log.info(f"destroying redis {self.resources['name']}")
        except HttpError as e:
            if e.resp.status == 404:
                log.info(f"redis {self.resources['name']} already destroyed")
            else:
                raise e

    def get(self):
        if not self.resources:
            return
        resp = self.versioned_clients.redis.execute(
            "get",
            parent_key="name",
            parent_schema="projects/{project_id}/locations/{location_id}/instances/"
            + self.resources["name"],
        )
        return resp

    def get_config(self):
        if not self.resources:
            return
        redis = self.get()
        return {
            "resource_type": self.resource_type,
            "values": {
                "REDIS_INSTANCE_NAME": redis["name"],
                "REDIS_HOST": redis["host"],
                "REDIS_PORT": f"{redis['port']}",
            },
        }



import re
from os.path import join, dirname

DATA_DIR = join(dirname(__file__), "tests", "data", "http")
PROJECT_ID = "goblet"
DATA_DIR_MAIN = join(dirname(__file__), "tests", "data")


def sanitize_project_name(dirty_str):
    sanitized = "projects/{}/".format(PROJECT_ID)
    return re.sub(r"projects/([0-9a-zA-Z_-]+)/", sanitized, dirty_str)


def dummy_function():
    return None


def mock_dummy_function(mock):
    def dummy_function(event=None):
        return mock()

    return dummy_function



import logging
import os

from goblet.handlers.handler import Handler
from goblet.permissions import gcp_generic_resource_permissions
from goblet.utils import nested_update
from goblet_gcp_client.client import get_default_project, get_default_location
from googleapiclient.errors import HttpError

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class Uptime(Handler):
    """Uptime Trigger
    Uptime checks only support public https urls and cloud run revisions currently https://cloud.google.com/monitoring/uptime-checks
    """

    resource_type = "uptime"
    valid_backends = ["cloudfunction", "cloudfunctionv2", "cloudrun"]
    required_apis = ["monitoring"]
    permissions = [
        *gcp_generic_resource_permissions("monitoring", "uptimeCheckConfigs")
    ]

    def __init__(
        self, name, backend, versioned_clients=None, resources=None, routes_type=None
    ):
        super(Uptime, self).__init__(
            name=name,
            versioned_clients=versioned_clients,
            resources=resources,
            backend=backend,
        )
        self.resources = resources or {}
        self.routes_type = routes_type

    def register(self, name, func, kwargs):
        self.resources[name] = {"func": func, "name": name, "kwargs": kwargs["kwargs"]}

    def __call__(self, request, context=None):
        headers = request.headers or {}
        uptime = self.resources[headers["X-Goblet-Uptime-Name"]]
        return uptime["func"]()

    def _deploy(self, source=None, entrypoint=None):
        # TODO: support api gateway

        if self.resources:
            checks = self.list_uptime_checks()
        for name, uptime in self.resources.items():
            uptime_config = {"displayName": f"{self.name}-{name}"}
            if (
                self.backend.resource_type == "cloudrun"
                and not self.routes_type == "apigateway"
            ):
                uptime_config["monitoredResource"] = {
                    "type": "cloud_run_revision",
                    "labels": {
                        "location": "us-central1",
                        "service_name": self.name,
                        "revision_name": "",
                        "configuration_name": "",
                        "project_id": get_default_project(),
                    },
                }
                uptime_config["httpCheck"] = {
                    "useSsl": True,
                    "headers": {"X-Goblet-Uptime-Name": uptime["name"]},
                }
            if self.backend.resource_type.startswith("cloudfunction"):
                uptime_config["monitoredResource"] = {
                    "type": "uptime_url",
                    "labels": {
                        "host": f"{get_default_location()}-{get_default_project()}.cloudfunctions.net",
                        "project_id": get_default_project(),
                    },
                }
                uptime_config["httpCheck"] = {
                    "useSsl": True,
                    "path": f"/{self.name}",
                    "headers": {"X-Goblet-Uptime-Name": uptime["name"]},
                }

            # update config based on user arguments
            uptime_config = nested_update(uptime_config, uptime["kwargs"])

            # Uptime exists
            check = [
                check
                for check in checks
                if check["displayName"] == uptime_config["displayName"]
            ]
            if len(check) == 1:
                # Setup update mask
                keys = list(uptime_config.keys())

                # Remove keys that cannot be updated
                keys.remove("monitoredResource")
                updateMask = ",".join(keys)

                self.versioned_clients.monitoring_uptime.execute(
                    "patch",
                    parent_key="name",
                    parent_schema=check[0]["name"],
                    params={"body": uptime_config, "updateMask": updateMask},
                )
                log.info(f"updated uptime check: {name} for {self.name}")

            else:
                self.versioned_clients.monitoring_uptime.execute(
                    "create", params={"body": uptime_config}
                )
                log.info(f"created uptime check: {name} for {self.name}")

        return

    def destroy(self):
        if not self.resources:
            return
        for check in self.list_uptime_checks():
            self._destroy_uptime_check(check)

    def _destroy_uptime_check(self, check):
        try:
            self.versioned_clients.monitoring_uptime.execute(
                "delete", parent_key="name", parent_schema=check["name"]
            )
            log.info(f"Destroying uptime check {check['displayName']}......")
        except HttpError as e:
            if e.resp.status == 404:
                log.info("Uptime check already destroyed")
            else:
                raise e

    def list_uptime_checks(self):
        resp = self.versioned_clients.monitoring_uptime.execute(
            "list",
            parent_key="parent",
            params={"filter": f"displayName=starts_with('{self.name}-')"},
        )
        return resp.get("uptimeCheckConfigs", [])

    def set_invoker_permissions(self):
        return



import json


class Response(object):
    """
    Generic Response class based on Flask Response
    """

    def __init__(self, body: str, headers=None, status_code=200):
        self.body = body
        if headers is None:
            headers = {"Content-type": "text/plain"}
        self.headers = headers
        self.status_code = status_code

    def __call__(self, environ, start_response):
        body = self.body
        if not isinstance(body, (str, bytes)):
            body = json.dumps(body, separators=(",", ":"))
        status = self.status_code
        headers = [(k, v) for k, v in self.headers.items()]
        start_response(status, headers)
        return [body]


def default_missing_route(error):
    return Response(str(error), status_code=404)



import logging
import os

from goblet.common_cloud_actions import deploy_apigateway, destroy_apigateway
from goblet.infrastructures.infrastructure import Infrastructure
from goblet.handlers.routes import OpenApiSpec
from goblet.utils import get_g_dir, get_dir
from goblet.permissions import gcp_generic_resource_permissions

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class ApiGateway(Infrastructure):
    """Api Gateway that is deployed with an existing openapi spec"""

    resource_type = "apigateway"
    required_apis = ["apigateway"]
    permissions = [
        "apigateway.operations.get",
        *gcp_generic_resource_permissions("apigateway", "apiconfigs"),
        *gcp_generic_resource_permissions("apigateway", "apis"),
        *gcp_generic_resource_permissions("apigateway", "gateways"),
    ]

    def register(self, name, **kwargs):
        kwargs = kwargs["kwargs"]
        self.resources = {
            "name": name,
            "backend_url": kwargs["backend_url"],
            "openapi_dict": kwargs["openapi_dict"],
        }

    def _deploy(self):
        if not self.resources:
            return
        goblet_spec = OpenApiSpec(
            self.resources["name"],
            self.resources["backend_url"],
            existing_spec=self.resources["openapi_dict"],
        )
        goblet_spec.add_x_google_backend()

        # create .goblet if doesnt exist
        if not os.path.isdir(f"{get_dir()}/.goblet"):
            os.mkdir(f"{get_dir()}/.goblet")

        updated_filename = f"{get_g_dir()}/{self.resources['name']}_openapi_spec.yml"
        with open(updated_filename, "w") as f:
            goblet_spec.write(f)
        deploy_apigateway(
            self.resources["name"],
            self.config,
            self.versioned_clients,
            updated_filename,
        )

    def destroy(self):
        if not self.resources:
            return
        destroy_apigateway(self.resources["name"], self.versioned_clients)



from googleapiclient.errors import HttpError
from goblet.infrastructures.infrastructure import Infrastructure
from goblet.permissions import gcp_generic_resource_permissions

import logging
import os

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class VPCConnector(Infrastructure):
    resource_type = "vpcconnector"
    required_apis = ["vpcaccess"]
    permissions = [
        "vpcaccess.operations.get",
        gcp_generic_resource_permissions("vpcaccess", "connectors"),
    ]

    def register(self, name, kwargs):
        self.resources = {"name": name}
        vpcconnector_config = self.config.vpcconnector or {}

        if not vpcconnector_config.get("ipCidrRange"):
            raise ValueError("ipCidrRange not specified in config")

    def _deploy(self):
        if not self.resources:
            return
        vpcconnector_config = self.config.vpcconnector or {}

        # either min/max throughput or instances needs to be set
        req_body = {
            "network": vpcconnector_config.get("network", "default"),
            "ipCidrRange": vpcconnector_config.get("ipCidrRange"),
            "minInstances": vpcconnector_config.get("minInstances", 2),  # DEFAULT
            "maxInstances": vpcconnector_config.get(
                "maxInstances", vpcconnector_config.get("minInstances", 2) + 1
            ),
            **vpcconnector_config,
        }

        try:
            resp = self.versioned_clients.vpcconnector.execute(
                "create",
                params={"connectorId": self.resources["name"], "body": req_body},
            )
            self.versioned_clients.vpcconnector.wait_for_operation(resp["name"])
        except HttpError as e:
            if e.resp.status == 409:
                log.info(
                    f"vpc connector {self.resources['name']} already exists, updating not supported"
                )
                pass
            else:
                raise e

    def get(self):
        if not self.resources:
            return
        resp = self.versioned_clients.vpcconnector.execute(
            "get",
            parent_key="name",
            parent_schema="projects/{project_id}/locations/{location_id}/connectors/"
            + self.resources["name"],
        )
        return resp

    def destroy(self):
        if not self.resources:
            return
        try:
            resp = self.versioned_clients.vpcconnector.execute(
                "delete",
                parent_key="name",
                parent_schema="projects/{project_id}/locations/{location_id}/connectors/"
                + self.resources["name"],
            )
            self.versioned_clients.vpcconnector.wait_for_operation(resp["name"])
            log.info("destroying vpc connector")
        except HttpError as e:
            if e.resp.status == 404:
                log.info(f"vpc connector {self.resources['name']} already deleted")
            else:
                raise e

    def get_config(self):
        if not self.resources:
            return
        vpc_connector = self.get()

        return {
            "resource_type": self.resource_type,
            "values": {
                "name": vpc_connector["name"],
                "egress": "PRIVATE_RANGES_ONLY",
            },
        }



import logging
import os

from goblet.handlers.handler import Handler
from goblet.common_cloud_actions import get_artifact_image_name
from goblet.permissions import gcp_generic_resource_permissions

from googleapiclient.errors import HttpError

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class Jobs(Handler):
    """Cloudrun job
    https://cloud.google.com/run/docs/create-jobs
    """

    resource_type = "job"
    valid_backends = ["cloudrun"]
    can_sync = True
    required_apis = ["cloudbuild", "run", "cloudfunctions"]
    permissions = [
        *gcp_generic_resource_permissions("run", "jobs"),
        "run.operations.get",
        "cloudbuild.builds.create",
        "cloudbuild.builds.get",
        "cloudbuild.builds.list",
        "cloudfunctions.functions.sourceCodeSet",
    ]

    def register(self, name, func, kwargs):
        task_id = kwargs["task_id"]
        job_name = f"{self.name}-{kwargs['name']}"
        if self.resources.get(job_name):
            self.resources[job_name][task_id] = {"func": func}
        else:
            self.resources[job_name] = {
                task_id: {"func": func},
                "execution_spec": kwargs.get("kwargs", {}),
            }

    def __call__(self, name, task_id):
        if not self.resources.get(name):
            raise ValueError(f"Job {name} not found")

        if not self.resources[name].get(task_id):
            raise ValueError(
                f"Job {name} not found for CLOUD_RUN_TASK_INDEX: {task_id}"
            )

        job = self.resources[name][task_id]

        return job["func"](task_id)

    def _deploy(self, source=None, entrypoint=None):
        if not self.resources:
            return

        artifact = get_artifact_image_name(
            self.versioned_clients.cloudbuild, self.name, config=self.config
        )

        log.info("deploying cloudrun jobs......")
        for job_name, job in self.resources.items():
            container = {**(self.config.job_container or {})}
            container["image"] = artifact
            container["command"] = [
                "goblet",
                "job",
                "run",
                job_name,
            ]

            job_spec = {
                "launchStage": "BETA",
                "labels": self.config.labels,
                "template": {
                    "taskCount": len(job.keys()) - 1,
                    "template": {
                        "containers": [container],
                        **(self.config.job_spec or {}),
                    },
                    **job["execution_spec"],
                },
            }

            self.deploy_job(job_name, job_spec)

    def _sync(self, dryrun=False):
        jobs = self.versioned_clients.run_job.execute("list").get("jobs", [])
        filtered_jobs = list(filter(lambda job: self.name in job["name"], jobs))
        for filtered_job in filtered_jobs:
            filtered_name = filtered_job["name"].split("jobs/")[-1]
            if not self.resources.get(filtered_name):
                log.info(f"Detected unused job in GCP {filtered_name}")
                if not dryrun:
                    self._destroy_job(filtered_name)

    def deploy_job(self, job_name, job):
        try:
            resp = self.versioned_clients.run_job.execute(
                "create", params={"jobId": job_name, "body": job}
            )
            self.versioned_clients.run_job.wait_for_operation(resp["name"])
            log.info(f"created job: {job_name}")
        except HttpError as e:
            if e.resp.status == 409:
                resp = self.versioned_clients.run_job.execute(
                    "patch",
                    parent_key="name",
                    parent_schema="projects/{project_id}/locations/{location_id}/jobs/"
                    + job_name,
                    params={"body": job},
                )
                self.versioned_clients.run_job.wait_for_operation(resp["name"])
                log.info(f"updated job: {job_name}")
            else:
                raise e

    def destroy(self):
        if not self.resources:
            return
        for job_name in self.resources.keys():
            self._destroy_job(job_name)

    def _destroy_job(self, job_name):
        try:
            resp = self.versioned_clients.run_job.execute(
                "delete",
                parent_key="name",
                parent_schema="projects/{project_id}/locations/{location_id}/jobs/"
                + job_name,
            )
            self.versioned_clients.run_job.wait_for_operation(resp["name"])
            log.info(f"Destroying job {job_name}......")
        except HttpError as e:
            if e.resp.status == 404:
                log.info("Jobs already destroyed")
            else:
                raise e

    def set_invoker_permissions(self):
        if self.config.bindings:
            for job_name in self.resources.keys():
                log.info(f"adding IAM bindings for cloudrun job {job_name}")
                policy_bindings = {"policy": {"bindings": self.config.bindings}}
                self.versioned_clients.run_job.execute(
                    "setIamPolicy",
                    parent_key="resource",
                    parent_schema="projects/{project_id}/locations/{location_id}/jobs/"
                    + job_name,
                    params={"body": policy_bindings},
                )



import google_auth_httplib2
import logging
import json
from urllib.parse import quote_plus
import base64
from googleapiclient.errors import HttpError

import goblet.globals as g
import os
from goblet.client import (
    VersionedClients,
    get_default_project_number,
)
from goblet_gcp_client.client import (
    Client,
    get_default_location,
    get_credentials,
    get_default_project,
)
from goblet.errors import GobletError
from goblet.utils import get_python_runtime
from goblet.permissions import add_binding
from typing import List

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


def check_or_enable_service(resources: List[str], enable: bool = False):
    if len(resources) == 0:
        return
    client = VersionedClients().service_usage
    if enable:
        resp = client.execute(
            "batchEnable",
            parent_key="parent",
            parent_schema="projects/{project_id}",
            params={
                "body": {
                    "serviceIds": [
                        f"{resource}.googleapis.com" for resource in resources
                    ]
                }
            },
        )
        if not resp.get("done"):
            client.wait_for_operation(resp["name"], calls="operations")
        for resource in resources:
            log.info(f"{resource} enabled")
    else:
        resp = client.execute(
            "batchGet",
            parent_key="parent",
            parent_schema="projects/{project_id}",
            params={
                "names": [
                    f"projects/{get_default_project()}/services/{resource}.googleapis.com"
                    for resource in resources
                ]
            },
        )
        for service in resp["services"]:
            log.info(f"{service['config']['name']} {service['state']}")
    return None


####### Cloud Functions #######
def create_cloudfunctionv1(client: Client, params: dict, config=None):
    create_cloudfunction(
        client, params, parent_key="location", operations_calls="operations"
    )


def create_cloudfunctionv2(client: Client, params: dict):
    create_cloudfunction(
        client,
        params,
        parent_key="parent",
        operations_calls="projects.locations.operations",
    )


def create_cloudfunction(
    client: Client,
    params: dict,
    parent_key="location",
    operations_calls="operations",
):
    """Creates a cloudfunction based on req_body"""
    function_name = params["body"]["name"].split("/")[-1]
    try:
        resp = client.execute("create", parent_key=parent_key, params=params)
        log.info(f"creating cloudfunction {function_name}")
    except HttpError as e:
        if e.resp.status == 409:
            log.info(f"updating cloudfunction {function_name}")
            resp = client.execute(
                "patch",
                parent_key="name",
                parent_schema=params["body"]["name"],
                params={"body": params["body"]},
            )
        else:
            raise e

    client.wait_for_operation(resp["name"], calls=operations_calls)

    # Set IAM Bindings
    if g.config.bindings:
        log.info(f"adding IAM bindings for cloudfunction {function_name}")
        policy_bindings = {"policy": {"bindings": g.config.bindings}}
        resp = client.execute(
            "setIamPolicy",
            parent_key="resource",
            parent_schema=params["body"]["name"],
            params={"body": policy_bindings},
        )


def destroy_cloudfunction(client, name):
    """Destroys cloudfunction"""
    try:
        client.execute(
            "delete",
            parent_schema="projects/{project_id}/locations/{location_id}/functions/"
            + name,
            parent_key="name",
        )
        log.info(f"deleting google cloudfunction {name}......")
    except HttpError as e:
        if e.resp.status == 404:
            log.info(f"cloudfunction {name} already destroyed")
        else:
            raise e


def destroy_cloudfunction_artifacts(name):
    """Destroys all images stored in cloud storage that are related to the function."""
    client = Client("cloudresourcemanager", "v1", calls="projects")
    resp = client.execute(
        "get", parent_key="projectId", parent_schema=get_default_project()
    )
    project_number = resp["projectNumber"]
    region = get_default_location()
    if not region:
        raise Exception("Missing Region")
    bucket_name = f"gcf-sources-{project_number}-{get_default_location()}"
    http = client.http or google_auth_httplib2.AuthorizedHttp(get_credentials())
    resp = http.request(
        f"https://storage.googleapis.com/storage/v1/b/{bucket_name}/o?prefix={name}"
    )
    objects = json.loads(resp[1])
    if not objects.get("items"):
        log.info("Artifacts already deleted")
        return
    for storage in objects["items"]:
        log.info(f"Deleting artifact {storage['name']}")
        resp = http.request(
            f"https://storage.googleapis.com/storage/v1/b/{bucket_name}/o/{quote_plus(storage['name'])}",
            method="DELETE",
        )


def get_cloudfunction_url(client, name):
    """Get the cloudrun url"""
    if client.version == "v1":
        return f"https://{get_default_location()}-{get_default_project()}.cloudfunctions.net/{name}"
    try:
        resp = client.execute(
            "get",
            parent_key="name",
            parent_schema="projects/{project_id}/locations/{location_id}/functions/"
            + name,
        )
        # handle both cases: first for gcf v1 and second for v2
        try:
            target = resp["httpsTrigger"]["url"]
        except KeyError:
            target = resp["serviceConfig"]["uri"]
        return target
    except HttpError as e:
        if e.resp.status == 404:
            log.info(f"cloudfunction {name} not found")
        else:
            raise e


def get_function_runtime(client, config=None):
    """
    Returns the proper runtime to be used in cloudfunctions
    """
    runtime = config.runtime or get_python_runtime()
    required_runtime = "python37" if client.version == "v1" else "python38"
    if int(runtime.split("python")[-1]) < int(required_runtime.split("python")[-1]):
        raise ValueError(
            f"Your current python runtime is {runtime}. Your backend requires a minimum of {required_runtime}"
            f". Either upgrade python on your machine or set the 'runtime' field in config.json."
        )
    return runtime


####### Cloud Run #######


def destroy_cloudrun(client, name):
    """Destroys cloudrun"""
    try:
        client.execute(
            "delete",
            parent_schema="projects/{project_id}/locations/{location_id}/services/"
            + name,
            parent_key="name",
        )
        log.info(f"deleting cloudrun {name}......")
    except HttpError as e:
        if e.resp.status == 404:
            log.info(f"cloudrun {name} already destroyed")
        else:
            raise e


def deploy_cloudrun(client, req_body, name):
    """Deploys cloud build to cloudrun"""
    try:
        params = {"body": req_body, "serviceId": name}
        resp = client.execute(
            "create",
            parent_key="parent",
            parent_schema="projects/"
            + get_default_project_number()
            + "/locations/{location_id}",
            params=params,
        )
        log.info("creating cloudrun")
    except HttpError as e:
        if e.resp.status == 409:
            log.info("updating cloudrun")
            resp = client.execute(
                "patch",
                parent_key="name",
                parent_schema="projects/"
                + get_default_project_number()
                + "/locations/{location_id}/services/"
                + name,
                params={"body": req_body},
            )
        else:
            raise e
    client.wait_for_operation(resp["name"], calls="projects.locations.operations")


def get_cloudrun_url(client, name):
    """Get the cloudrun url"""
    try:
        resp = client.execute(
            "get",
            parent_key="name",
            parent_schema="projects/{project_id}/locations/{location_id}/services/"
            + name,
        )
        # Handle both cases: cloudrun v1 and v2
        try:
            target = resp["status"]["url"]
        except KeyError:
            target = resp["uri"]
        return target
    except HttpError as e:
        if e.resp.status == 404:
            log.info(f"cloudrun {name} not found")
        else:
            raise e


####### Cloud Build #######


def create_cloudbuild(client, req_body):
    """Creates a cloudbuild based on req_body"""
    defaultProject = get_default_project()
    defaultLocation = "global"
    try:
        resp = client.execute(
            "create",
            parent_key="projectId",
            parent_schema=defaultProject,
            params={
                "body": req_body,
                "parent": f"projects/{defaultProject}/locations/{defaultLocation}",
            },
        )
        log.info("creating cloudbuild")
    except HttpError as e:
        raise e
    cloudbuild_config = g.config.cloudbuild or {}
    timeout_seconds = cloudbuild_config.get("timeout", "600s")
    if "s" not in timeout_seconds:
        log.info(
            "Not a valid timeout. Needs to be a duration that ends is 's'. Defaulting to 600s"
        )
        timeout = 600
    else:
        timeout = int(timeout_seconds.split("s")[0])
    resp = client.wait_for_operation(resp["name"], calls="operations", timeout=timeout)
    if not resp:
        raise GobletError("Build Timed out")
    if resp.get("error"):
        raise GobletError(
            f"Cloud build failed with error code {resp['error']['code']} and message {resp['error']['message']}"
        )


class MissingArtifact(Exception):
    """Raised when missing Cloudrun Artifact."""

    def __init__(self, missing):
        self.missing = missing


def getDefaultRegistry(artifactName):
    return f"{get_default_location()}-docker.pkg.dev/{get_default_project()}/cloud-run-source-deploy/{artifactName}"


def getDefaultRegistryName():
    return f"projects/{get_default_project()}/locations/{get_default_location()}/repositories/cloud-run-source-deploy"


def get_artifact_image_name(client, artifact_name, config):
    """retrieves name of image stored in artifact registry"""
    # Get artifact tag from env variable or config
    try:
        artifact_tag = os.environ["GOBLET_ARTIFACT_TAG"]
    except KeyError:
        artifact_tag = config.deploy.get("artifact_tag", None)

    if artifact_tag:
        artifact_registry = config.deploy.get(
            "artifact_registry", None
        ) or getDefaultRegistry(artifact_name)
        image_name = (
            artifact_registry
            + ("@" if "sha256" in artifact_tag else ":")
            + artifact_tag
        )
    else:
        image_name = getCloudbuildArtifact(client, artifact_name, config=config)
    return image_name


# calls latest build and checks for its artifact to avoid image:latest behavior with cloud run revisions
def getCloudbuildArtifact(client, artifactName, config):
    defaultProject = get_default_project()
    resp = client.execute(
        "list", parent_key="projectId", parent_schema=defaultProject, params={}
    )

    # search for latest build with artifactName
    latestArtifact = None
    try:
        registry = config.deploy.get("artifact_registry") or getDefaultRegistry(
            artifactName
        )
    except AttributeError:
        registry = getDefaultRegistry(artifactName)

    for build in resp["builds"]:
        # pending builds will not have results field.
        if (
            build.get("results")
            and build["results"].get("images")
            and registry == build["results"]["images"][0]["name"]
        ):
            latestArtifact = (
                build["results"]["images"][0]["name"]
                + "@"
                + build["results"]["images"][0]["digest"]
            )
            break

    if not latestArtifact:
        raise MissingArtifact("Missing artifact. Cloud Build may have failed.")
    return latestArtifact


####### Pub Sub #######


def get_pubsub_subscription(client, sub_name, req_body):
    try:
        resp = client.execute(
            "get",
            parent_key="subscription",
            parent_schema="projects/{project_id}/subscriptions/" + sub_name,
        )
    except HttpError as e:
        if e.resp.status == 404:
            return None
    return resp


def create_pubsub_subscription(client, sub_name, req_body, force_update=False):
    """Creates a pubsub subscription from req_body"""
    response = get_pubsub_subscription(client, sub_name, req_body)
    if response and response["topic"] != req_body["topic"]:
        log.info(
            f"Pubsub subscription projects do not match. {sub_name} is currently subscribed to {response['topic']}, but defined to be {req_body['topic']}."
        )
        if force_update:
            log.info("force_update is set to True. Deleting existing subscrition...")
            destroy_pubsub_subscription(client, sub_name)
            response = None
        else:
            log.info("force_update is set to False. Skipping update...")
            return

    # create pubsub
    if not response:
        client.execute(
            "create",
            parent_key="name",
            parent_schema="projects/{project_id}/subscriptions/" + sub_name,
            params={"body": req_body},
        )
        log.info(f"creating pubsub subscription {sub_name}")

    # update
    else:
        log.info(f"updating pubsub subscription {sub_name}")

        # Setup update mask
        keys = list(req_body.keys())
        # Remove keys that cannot be updated
        keys.remove("topic")
        if "filter" in keys:
            keys.remove("filter")
        if "labels" in keys:
            if req_body["labels"] is None or req_body["labels"] == {}:
                keys.remove("labels")
        if "enableMessageOrdering" in keys:
            keys.remove("enableMessageOrdering")
        updateMask = ",".join(keys)
        client.execute(
            "patch",
            parent_key="name",
            parent_schema="projects/{project_id}/subscriptions/" + sub_name,
            params={"body": {"subscription": req_body, "updateMask": updateMask}},
        )


def destroy_pubsub_subscription(client, name):
    """Destroys pubsub subscription"""
    try:
        client.execute(
            "delete",
            parent_key="subscription",
            parent_schema="projects/{project_id}/subscriptions/" + name,
        )
        log.info(f"deleting pubsub subscription {name}......")
    except HttpError as e:
        if e.resp.status == 404:
            log.info(f"pubsub subscription {name} already destroyed")
        else:
            raise e


####### Event Arc #######


def create_eventarc_trigger(client, trigger_name, region, req_body):
    """Creates an eventarc trigger from req_body"""
    try:
        client.execute(
            "create",
            parent_key="parent",
            parent_schema="projects/{project_id}/locations/" + region,
            params={"body": req_body, "triggerId": trigger_name, "validateOnly": False},
        )
        log.info(f"creating eventarc trigger {trigger_name}")
    except HttpError as e:
        if e.resp.status == 409:
            log.info(f"updating eventarc trigger  {trigger_name}")
            # Setup update mask
            keys = list(req_body.keys())
            keys.remove("name")
            if "transport" in keys:
                keys.remove("transport")
            updateMask = ",".join(keys)
            client.execute(
                "patch",
                parent_key="name",
                parent_schema="projects/{project_id}/locations/"
                + region
                + "/triggers/"
                + trigger_name,
                params={"body": req_body, "updateMask": updateMask},
            )
        else:
            raise e


def destroy_eventarc_trigger(client, trigger_name, region):
    """Destroys eventarc trigger"""
    try:
        client.execute(
            "delete",
            parent_key="name",
            parent_schema="projects/{project_id}/locations/"
            + region
            + "/triggers/"
            + trigger_name,
        )
        log.info(f"deleting eventarc trigger  {trigger_name}......")
    except HttpError as e:
        if e.resp.status == 404:
            log.info(f"eventarc trigger  {trigger_name} already destroyed")
        else:
            raise e


####### Api Gateway #######


def deploy_apigateway(name, gconfig, versioned_clients, spec_path):
    "Deploy an Api, Api Gateway, & Api Config"
    try:
        resp = versioned_clients.apigateway_api.execute(
            "create",
            params={"apiId": name, "body": {"labels": gconfig.labels}},
        )
        versioned_clients.apigateway_api.wait_for_operation(resp["name"])
    except HttpError as e:
        if e.resp.status == 409:
            log.info("api already deployed")
        else:
            raise e
    config = {
        "openapiDocuments": [
            {
                "document": {
                    "path": spec_path,
                    "contents": base64.b64encode(open(spec_path, "rb").read()).decode(
                        "utf-8"
                    ),
                }
            }
        ],
        "labels": gconfig.labels,
        **(gconfig.apiConfig or {}),
    }
    try:
        config_version_name = name
        versioned_clients.apigateway_configs.execute(
            "create",
            params={"apiConfigId": name, "body": config},
            parent_schema="projects/{project_id}/locations/global/apis/" + name,
        )
    except HttpError as e:
        if e.resp.status == 409:
            log.info("updating api endpoints")
            configs = versioned_clients.apigateway_configs.execute(
                "list",
                parent_schema="projects/{project_id}/locations/global/apis/" + name,
            )
            # TODO: use hash
            version = len(configs["apiConfigs"])
            config_version_name = f"{name}-v{version}"
            versioned_clients.apigateway_configs.execute(
                "create",
                parent_schema="projects/{project_id}/locations/global/apis/" + name,
                params={"apiConfigId": config_version_name, "body": config},
            )
        else:
            raise e
    gateway = {
        "apiConfig": f"projects/{get_default_project()}/locations/global/apis/{name}/configs/{config_version_name}",
        "labels": gconfig.labels,
    }
    try:
        gateway_resp = versioned_clients.apigateway.execute(
            "create", params={"gatewayId": name, "body": gateway}
        )
    except HttpError as e:
        if e.resp.status == 409:
            log.info("updating gateway")
            gateway_resp = versioned_clients.apigateway.execute(
                "patch",
                parent_key="name",
                parent_schema="projects/{project_id}/locations/{location_id}/gateways/"
                + name,
                params={"updateMask": "apiConfig", "body": gateway},
            )
        else:
            raise e
    if gateway_resp:
        versioned_clients.apigateway.wait_for_operation(gateway_resp["name"])
    log.info("api successfully deployed...")
    gateway_resp = versioned_clients.apigateway.execute(
        "get",
        parent_key="name",
        parent_schema="projects/{project_id}/locations/{location_id}/gateways/" + name,
    )
    log.info(f"api endpoint is {gateway_resp['defaultHostname']}")


def destroy_apigateway(name, versioned_clients):
    """Destroy api gateway"""
    try:
        resp = versioned_clients.apigateway.execute(
            "delete",
            parent_schema="projects/{project_id}/locations/{location_id}/gateways/"
            + name,
            parent_key="name",
        )
        log.info("destroying api gateway......")
        versioned_clients.apigateway_configs.wait_for_operation(resp["name"])
    except HttpError as e:
        if e.resp.status == 404:
            log.info("api gateway already destroyed")
        else:
            raise e
    # destroy api config
    try:
        configs = versioned_clients.apigateway_configs.execute(
            "list",
            parent_schema="projects/{project_id}/locations/global/apis/" + name,
        )
        resp = {}
        log.info("api configs destroying....")
        for c in configs.get("apiConfigs", []):
            resp = versioned_clients.apigateway_configs.execute(
                "delete",
                parent_key="name",
                parent_schema="projects/{project_id}/locations/global/apis/"
                + name
                + "/configs/"
                + c["displayName"],
            )
            if resp:
                versioned_clients.apigateway_configs.wait_for_operation(resp["name"])
    except HttpError as e:
        if e.resp.status == 404:
            log.info("api configs already destroyed")
        else:
            raise e

    # destroy api
    try:
        resp = versioned_clients.apigateway_api.execute(
            "delete",
            parent_key="name",
            parent_schema="projects/{project_id}/locations/global/apis/" + name,
        )
        versioned_clients.apigateway_configs.wait_for_operation(resp["name"])
        log.info("apis successfully destroyed......")
    except HttpError as e:
        if e.resp.status == 404:
            log.info("api already destroyed")
        else:
            raise e


####### IAM #######


def deploy_custom_role(client, role):
    """Deploys a custom iam role"""
    try:
        params = {"body": role}
        resp = client.execute(
            "create",
            parent_key="parent",
            parent_schema="projects/" + get_default_project(),
            params=params,
        )
        log.info(f"creating custom iam role {role['roleId']}...")
    except HttpError as e:
        if e.resp.status == 409:
            log.info(f"updating custom iam role {role['roleId']}...")
            resp = client.execute(
                "patch",
                parent_key="name",
                parent_schema="projects/"
                + get_default_project()
                + "/roles/"
                + role["roleId"],
                params={"body": role["role"], "updateMask": "includedPermissions"},
            )
        else:
            raise e
    return resp


def deploy_service_account(versioned_client, name, roleName):
    """Deploys service account with custom role.
    Service Account name needs to be [a-zA-Z][a-zA-Z-]*[a-zA-Z]
    """
    try:
        params = {
            "body": {
                "accountId": name,
                "serviceAccount": {
                    "displayName": name,
                    "description": "generated by goblet",
                },
            }
        }
        versioned_client.service_account.execute(
            "create",
            parent_key="name",
            parent_schema="projects/" + get_default_project(),
            params=params,
        )
        log.info(f"creating service account {name}...")
    except HttpError as e:
        if e.resp.status == 409:
            log.info(f"service account {name} already exists...")
        else:
            raise e
    add_binding(
        versioned_client.project_resource_manager,
        "projects/" + get_default_project(),
        f"projects/{get_default_project()}/roles/{roleName}",
        [f"serviceAccount:{name}@{get_default_project()}.iam.gserviceaccount.com"],
    )



from goblet.backends.cloudrun import CloudRun  # noqa: F401
from goblet.backends.cloudfunctionv1 import CloudFunctionV1  # noqa: F401
from goblet.backends.cloudfunctionv2 import CloudFunctionV2  # noqa: F401



import logging
import os

from goblet.infrastructures.infrastructure import Infrastructure
from goblet_gcp_client.client import get_default_project
from goblet.permissions import gcp_generic_resource_permissions

from googleapiclient.errors import HttpError

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class Alerts(Infrastructure):
    """Cloud Monitoring Alert Policies that can trigger notification channels based on built in or custom metrics.
    https://cloud.google.com/monitoring/api/ref_v3/rest/v3/projects.alertPolicies. Alerts and Alert conditions contain a
    few common defaults, that are used by GCP. These can be overriden by passing in the correct params.
    """

    resource_type = "alerts"
    can_sync = True
    _gcp_deployed_alerts = {}
    required_apis = ["logging"]
    permissions = [
        *gcp_generic_resource_permissions("monitoring", "alertPolicies"),
        *gcp_generic_resource_permissions("logging", "logMetrics"),
    ]

    def register(self, name, **kwargs):
        kwargs = kwargs.get("kwargs", {})
        conditions = kwargs.pop("conditions")
        notification_channels = kwargs.pop("notification_channels", [])
        self.resources[f"{self.name}-{name}"] = {
            "name": f"{self.name}-{name}",
            "conditions": conditions,
            "notification_channels": notification_channels,
            "kwargs": kwargs,
        }

    @property
    def gcp_deployed_alerts(self):
        """
        List of deployed gcp alerts, since we need to get unique id's from alerts in order to patch or avoid creating duplicates
        """
        if not self._gcp_deployed_alerts:
            alerts = self.versioned_clients.monitoring_alert.execute(
                "list",
                parent_key="name",
                params={"filter": f'display_name=starts_with("{self.name}-")'},
            )
            for alert in alerts.get("alertPolicies", []):
                self._gcp_deployed_alerts[alert["displayName"]] = alert
        return self._gcp_deployed_alerts

    def _deploy(self, source=None, entrypoint=None):
        if not self.resources:
            return
        config_notification_channels = []
        if self.config.alerts:
            config_notification_channels = self.config.alerts.get(
                "notification_channels", []
            )

        log.info("deploying alerts......")
        for alert_name, alert in self.resources.items():
            self.deploy_alert(alert_name, alert, config_notification_channels)

    def deploy_alert(self, alert_name, alert, notification_channels):
        formatted_conditions = []
        default_alert_kwargs = {}
        for condition in alert["conditions"]:
            condition.format_filter_or_query(
                self.name,
                self.backend.monitoring_type,
                self.backend.name,
                self.backend.monitoring_label_key,
            )
            default_alert_kwargs.update(condition.default_alert_kwargs)
            formatted_conditions.append(condition.condition)

            # deploy custom metrics if needed
            condition.deploy_extra(self.versioned_clients)

        default_alert_kwargs.update(alert["kwargs"])
        alert["notification_channels"].extend(notification_channels)

        body = {
            "displayName": alert_name,
            "conditions": formatted_conditions,
            "notificationChannels": alert["notification_channels"],
            "combiner": "OR",
            **default_alert_kwargs,
        }
        # check if exists
        if alert_name in self.gcp_deployed_alerts:
            # patch
            self.versioned_clients.monitoring_alert.execute(
                "patch",
                parent_key="name",
                parent_schema=self.gcp_deployed_alerts[alert_name]["name"],
                params={"updateMask": ",".join(body.keys()), "body": body},
            )

            log.info(f"updated alert: {alert_name}")
        else:
            # deploy
            self.versioned_clients.monitoring_alert.execute(
                "create",
                parent_key="name",
                params={"body": body},
            )
            log.info(f"created alert: {alert_name}")

    def _sync(self, dryrun=False):
        # Does not sync custom metrics
        for alert_name in self.gcp_deployed_alerts.keys():
            if not self.resources.get(alert_name):
                log.info(f"Detected unused alert {alert_name}")
                if not dryrun:
                    self._destroy_alert(alert_name)

    def destroy(self):
        if not self.resources:
            return
        for alert_name in self.resources.keys():
            self._destroy_alert(alert_name)

    def _destroy_alert(self, alert_name):
        if not self.gcp_deployed_alerts.get(alert_name):
            log.info(f"Alert {alert_name} already destroyed")
        else:
            try:
                self.versioned_clients.monitoring_alert.execute(
                    "delete",
                    parent_key="name",
                    parent_schema=self.gcp_deployed_alerts[alert_name]["name"],
                )
                log.info(f"Destroying alert {alert_name}......")
            except HttpError as e:
                if e.resp.status == 404:
                    log.info(f"Alert {alert_name} already destroyed")
                else:
                    raise e
        for condition in self.resources.get(alert_name, {}).get("conditions", []):
            condition.format_filter_or_query(
                self.name,
                self.backend.monitoring_type,
                self.backend.name,
                self.backend.monitoring_label_key,
            )
            condition.destroy_extra(self.versioned_clients)


class AlertCondition:
    """Base class for Alert Conditions. Only one of type threshold, absense, log_match, or MQL can be specified per condition.
    The method format_filter_or_query is used to inject values from the backend into the filters and conditions. These can be injected into
    custom filters as well. Currently monitoring_type,resource_name, and monitoring_label_key are supported.
    """

    def __init__(
        self, name, threshold=None, absence=None, log_match=None, MQL=None
    ) -> None:
        self.name = name
        self.app_name = ""
        if [threshold, absence, log_match, MQL].count(None) != 3:
            raise ValueError("Exactly 1 condition option can be set")
        if threshold:
            self.condition_key = "conditionThreshold"
            self._condition = threshold
        if absence:
            self.condition_key = "conditionAbsent"
            self._condition = absence
        if log_match:
            self.condition_key = "conditionMatchedLog"
            self._condition = log_match
        if MQL:
            self.condition_key = "conditionMonitoringQueryLanguage"
            self._condition = MQL
        self.filter = self._condition.get("filter")
        # For MQL
        self.query = self._condition.get("query")
        self.default_alert_kwargs = {}

    @property
    def condition(self):
        if self._condition.get("filter"):
            self._condition["filter"] = self.filter
        if self._condition.get("query"):
            self._condition["query"] = self.query
        return {
            "displayName": f"{self.app_name}-{self.name}",
            self.condition_key: self._condition,
        }

    def format_filter_or_query(
        self, app_name, monitoring_type, resource_name, monitoring_label_key
    ):
        self.app_name = app_name
        self.filter = self.filter.format(
            monitoring_type=monitoring_type,
            resource_name=resource_name,
            monitoring_label_key=monitoring_label_key,
        )
        return

    def deploy_extra(self, versioned_clients):
        pass

    def destroy_extra(self, versioned_clients):
        pass


class MetricCondition(AlertCondition):
    """
    https://cloud.google.com/monitoring/api/ref_v3/rest/v3/projects.alertPolicies#MetricThreshold
    """

    def __init__(
        self, name, metric, value, duration="60s", comparison="COMPARISON_GT", **kwargs
    ) -> None:
        super().__init__(
            name=name,
            threshold={
                "filter": 'resource.type="{monitoring_type}" AND resource.labels.{monitoring_label_key}="{resource_name}" AND metric.type = "{metric}"'.format(
                    metric=metric,
                    monitoring_type="{monitoring_type}",
                    monitoring_label_key="{monitoring_label_key}",
                    resource_name="{resource_name}",
                ),
                "thresholdValue": value,
                "duration": duration,
                "comparison": comparison,
                "aggregations": [
                    {
                        "alignmentPeriod": "300s",
                        "crossSeriesReducer": "REDUCE_NONE",
                        "perSeriesAligner": "ALIGN_MEAN",
                    }
                ],
                **kwargs,
            },
        )


class CustomMetricCondition(MetricCondition):
    """
    Creates and deploys a custom metric specified by the user, before creating a generic metric condition.
    """

    def __init__(
        self, name, metric_filter, value, metric_descriptor={}, **kwargs
    ) -> None:
        self.metric_filter = (
            'resource.type="{monitoring_type}" resource.labels.{monitoring_label_key}="{resource_name}" '
            + metric_filter
        )

        super().__init__(
            name=name,
            metric=f"logging.googleapis.com/user/{{resource_name}}-{name}",
            value=value,
            **kwargs,
        )

    def format_filter_or_query(
        self, app_name, monitoring_type, resource_name, monitoring_label_key
    ):
        self.app_name = app_name
        self.filter = self.filter.format(
            monitoring_type=monitoring_type,
            resource_name=resource_name,
            monitoring_label_key=monitoring_label_key,
        )

        self.metric_filter = self.metric_filter.format(
            monitoring_type=monitoring_type,
            resource_name=resource_name,
            monitoring_label_key=monitoring_label_key,
        )
        return

    def deploy_extra(self, versioned_clients):
        # Defaults
        metric_descriptor = {
            "name": f"projects/{get_default_project}/metricDescriptors/logging.googleapis.com/user/{self.app_name}-{self.name}",
            "metricKind": "DELTA",
            "valueType": "INT64",
            "unit": "1",
            "description": "measure error rates in media-metadata-service cloudfunction",
            "type": f"logging.googleapis.com/user/{self.app_name}-{self.name}",
        }
        # User Override
        metric_descriptor.update(metric_descriptor)

        metric_body = {
            "name": f"{self.app_name}-{self.name}",
            "description": f"Goblet generated custom metric for metric {self.app_name}-{self.name}",
            "filter": self.metric_filter,
            "metricDescriptor": metric_descriptor,
        }
        try:
            versioned_clients.logging_metric.execute(
                "create",
                params={"body": metric_body},
            )
            log.info(f"deploying custom metric {self.app_name}-{self.name}")
        except HttpError as e:
            if e.resp.status == 409:
                log.info(f"updating custom metric {self.app_name}-{self.name}")
                versioned_clients.logging_metric.execute(
                    "update",
                    parent_key="metricName",
                    parent_schema=f"projects/{get_default_project()}/metrics/{self.app_name}-{self.name}",
                    params={"body": metric_body},
                )
            else:
                raise e

    def destroy_extra(self, versioned_clients):
        try:
            versioned_clients.logging_metric.execute(
                "delete",
                parent_key="metricName",
                parent_schema=f"projects/{get_default_project()}/metrics/{self.app_name}-{self.name}",
            )
            log.info(f"deleting custom metric {self.app_name}-{self.name}")
        except HttpError as e:
            if e.resp.status == 404:
                log.info(f"Custom metric {self.app_name}-{self.name} already destroyed")
            else:
                raise e


class LogMatchCondition(AlertCondition):
    """
    https://cloud.google.com/monitoring/api/ref_v3/rest/v3/projects.alertPolicies#LogMatch
    """

    def __init__(self, name, filter, **kwargs) -> None:
        super().__init__(
            name=name,
            log_match={
                "filter": filter
                if kwargs.get("replace_filter", False)
                else 'resource.type="{monitoring_type}"\nresource.labels.{monitoring_label_key}="{resource_name}"\n'
                + filter
            },
        )
        self.default_alert_kwargs = {
            "alertStrategy": {
                "notificationRateLimit": {"period": "300s"},
                "autoClose": "604800s",
            }
        }


class PubSubDLQCondition(MetricCondition):
    """
    Creates and deploys an alert for dead letter queue messages in a pubsub subscription.
    """

    def __init__(self, name, subscription_id, value=0, **kwargs) -> None:
        super().__init__(
            name=name,
            metric="pubsub.googleapis.com/subscription/dead_letter_message_count",
            value=value,
            filter='resource.labels.subscription_id = "{subscription_id}" AND resource.type = "pubsub_subscription" AND metric.type = "pubsub.googleapis.com/subscription/dead_letter_message_count"'.format(
                subscription_id=subscription_id
            ),
            **kwargs,
        )


class UptimeCondition(MetricCondition):
    """
    Creates and deploys an alert for failed uptime checks.
    Supports `uptime_url` or `cloud_run_revision`
    """

    def __init__(self, name, check_id, resource_type, value=1, **kwargs) -> None:
        super().__init__(
            name=name,
            metric="monitoring.googleapis.com/uptime_check/check_passed",
            value=value,
            filter='metric.labels.check_id = "{check_id}" AND resource.type = "{resource_type}" AND metric.type = "monitoring.googleapis.com/uptime_check/check_passed"'.format(
                check_id=check_id, resource_type=resource_type
            ),
            aggregations=[
                {
                    "alignmentPeriod": "1200s",
                    "crossSeriesReducer": "REDUCE_COUNT_FALSE",
                }
            ],
            **kwargs,
        )



from requests import request
import base64

from goblet.backends.backend import Backend
from goblet.client import VersionedClients
from goblet_gcp_client.client import get_default_location, get_default_project
from goblet.common_cloud_actions import (
    get_function_runtime,
    create_cloudfunctionv1,
    destroy_cloudfunction_artifacts,
    destroy_cloudfunction,
)
from goblet.permissions import gcp_generic_resource_permissions, add_binding


class CloudFunctionV1(Backend):
    resource_type = "cloudfunction"
    supported_versions = ["v1"]
    required_files = ["requirements.txt", "main.py"]
    config_key = "cloudfunction"
    monitoring_type = "cloud_function"
    monitoring_label_key = "function_name"
    required_apis = [
        "cloudfunctions",
        "secretmanager",
        "cloudresourcemanager",
        "cloudbuild",
    ]
    permissions = [
        "cloudfunctions.functions.getIamPolicy",
        "cloudfunctions.functions.setIamPolicy",
        "cloudfunctions.operations.get",
        "cloudfunctions.functions.sourceCodeSet",
        "cloudresourcemanager.projects.get",
        "iam.serviceaccounts.actAs",
        *gcp_generic_resource_permissions("cloudfunctions", "functions"),
    ]

    def __init__(self, app):
        self.client = VersionedClients().cloudfunctions
        self.func_path = f"projects/{get_default_project()}/locations/{get_default_location()}/functions/{app.function_name}"
        super().__init__(app, self.client, self.func_path)

    def deploy(self, force=False):
        put_headers = {
            "content-type": "application/zip",
            "x-goog-content-length-range": "0,104857600",
        }
        source, changes = self._gcs_upload(self.client, put_headers, force=force)
        if not changes:
            return None

        if self.app.is_http():
            client, params = self._get_upload_params(source)
            create_cloudfunctionv1(client, params, config=self.config)

        return source

    def destroy(self, all=False):
        destroy_cloudfunction(self.client, self.name)
        if all:
            destroy_cloudfunction_artifacts(self.name)

    def _get_upload_params(self, source):
        user_configs = self.config.cloudfunction or {}
        params = {
            "body": {
                "name": self.func_path,
                "description": self.config.description or "created by goblet",
                "entryPoint": "goblet_entrypoint",
                "sourceUploadUrl": source["uploadUrl"],
                "httpsTrigger": {},
                "runtime": get_function_runtime(self.client, self.config),
                "labels": {**self.config.labels},
                **user_configs,
            }
        }
        return self.client, params

    def _checksum(self):
        source_info = self.client.execute(
            "generateDownloadUrl", parent_key="name", parent_schema=self.func_path
        )
        resp = request("HEAD", source_info["downloadUrl"])
        return resp.headers["x-goog-hash"].split(",")[-1].split("=", 1)[-1]

    def update_config(self, infra_configs=[], write_config=False, stage=None):
        config_updates = {self.config_key: {}}
        for infra_config in infra_configs:
            resource_type = infra_config["resource_type"]
            if resource_type == "vpcconnector":
                config_updates[self.config_key] = {
                    **config_updates.get(self.config_key, {}),
                    "vpcConnector": infra_config["values"]["name"],
                    "vpcConnectorEgressSettings": infra_config["values"]["egress"],
                }

            elif resource_type in ("redis", "cloudtaskqueue"):
                config_updates[self.config_key]["environmentVariables"] = {
                    **config_updates[self.config_key].get("environmentVariables", {}),
                    **infra_config.get("values"),
                }

        self.config.update_g_config(
            values=config_updates, write_config=write_config, stage=stage
        )

    @property
    def http_endpoint(self):
        return f"https://{get_default_location()}-{get_default_project()}.cloudfunctions.net/{self.name}"

    # TODO:
    def set_iam_policy(self, service_account_id):
        client = self.client
        resource_name = self.func_path
        policy = {
            "policy": {
                "bindings": {
                    "role": "roles/cloudfunctions.invoker",
                    "members": [f"serviceAccount:{service_account_id}"],
                }
            }
        }
        client.execute(
            "setIamPolicy",
            params={"body": policy},
            parent_key="resource",
            parent_schema=resource_name,
        )

    def get_environment_vars(self):
        env_dict = self.config.config.get("cloudfunction", {}).get(
            "environmentVariables", {}
        )

        versioned_clients = VersionedClients()
        for secret in self.config.config.get("cloudfunction", {}).get(
            "secretEnvironmentVariables", []
        ):
            try:
                secret_name = secret["secret"]
                version = secret["version"]

                resp = versioned_clients.secretmanager.execute(
                    "access",
                    parent_key="name",
                    parent_schema="projects/{project_id}/secrets/"
                    + secret_name
                    + "/versions/"
                    + version,
                )
                env_dict[secret["key"]] = base64.b64decode(
                    resp["payload"]["data"]
                ).decode()
            except Exception as e:
                self.log.info(
                    f"Unable to get secret {secret['key']} and set environment variable with error {str(e)}. Skipping..."
                )

        return env_dict

    def add_invoker_binding(self, principles):
        add_binding(
            self.client, self.func_path, "roles/cloudfunctions.invoker", principles
        )



import json
from base64 import b64encode
from googleapiclient.errors import HttpError
from goblet.infrastructures.infrastructure import Infrastructure
from goblet.permissions import gcp_generic_resource_permissions
import logging
import os
import datetime
from google.protobuf import duration_pb2, timestamp_pb2
from goblet.client import VersionedClients

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class CloudTaskClient:
    def __init__(self, service_account, queue, backend):
        self.service_account = service_account
        self.queue = queue
        self.backend = backend

    def build_task(self, target, payload, in_seconds, task_name, deadline):
        # https://cloud.google.com/tasks/docs/reference/rest/v2/projects.locations.queues.tasks/create
        task = {
            "httpRequest": {
                "httpMethod": "POST",
                "headers": {
                    "X-Goblet-CloudTask-Target": target,
                },
                "url": self.backend.http_endpoint,
                "oidcToken": {
                    "serviceAccountEmail": self.service_account,
                    "audience": self.backend.http_endpoint,
                },
            }
        }

        if payload is not None:
            if isinstance(payload, dict):
                # Convert dict to JSON string
                payload = json.dumps(payload)
                # specify http content-type to application/json
                task["httpRequest"]["headers"]["Content-Type"] = "application/json"

            # The API expects a payload of type bytes.
            converted_payload = b64encode(payload.encode()).decode("utf8")
            # Add the payload to the request.
            task["httpRequest"]["body"] = converted_payload

        if in_seconds is not None:
            # Convert "seconds from now" into an rfc3339 datetime string.
            d = datetime.datetime.utcnow() + datetime.timedelta(seconds=in_seconds)

            # Create Timestamp protobuf.
            timestamp = timestamp_pb2.Timestamp()
            timestamp.FromDatetime(d)

            # Add the timestamp to the tasks.
            task["scheduleTime"] = timestamp.ToJsonString()

        if task_name is not None:
            # Add the name to tasks.
            task["name"] = f"{self.queue}/tasks/{task_name}"

        if deadline is not None:
            # Add dispatch deadline for requests sent to the worker.
            duration = duration_pb2.Duration()
            duration.FromSeconds(deadline)
            task["dispatchDeadline"] = duration.ToJsonString()

        return task

    def enqueue(self, target, payload, in_seconds=None, task_name=None, deadline=None):
        task = self.build_task(target, payload, in_seconds, task_name, deadline)
        return VersionedClients().cloudtask.execute(
            "create", parent_schema=self.queue, params={"body": {"task": task}}
        )


class CloudTaskQueue(Infrastructure):
    resource_type = "cloudtaskqueue"
    required_apis = ["cloudtasks"]
    permissions = gcp_generic_resource_permissions("cloudtasks", "queues")

    # https://cloud.google.com/apis/design/resource_names
    def register(self, name, kwargs):
        queue_config = kwargs.get("config", None)
        resource_id = name
        if (
            not queue_config
            and self.config.cloudtaskqueue
            and self.config.cloudtaskqueue.get(resource_id, None)
        ):
            queue_config = self.config.cloudtaskqueue.get(resource_id)
        self.resources[resource_id] = {
            "name": f"{self.versioned_clients.cloudtask_queue.parent}/queues/{resource_id}",
            "id": resource_id,
            "config": queue_config,
        }

        return CloudTaskClient(
            service_account=self.config.cloudtask.get("serviceAccount", None),
            queue=self.resources[resource_id]["name"],
            backend=self.backend,
        )

    def _deploy(self):
        if not self.resources:
            return

        for resource_id, resource in self.resources.items():
            params = {"name": resource["name"]}
            if resource["config"]:
                params = {**params, **resource["config"]}

            try:
                self.versioned_clients.cloudtask_queue.execute(
                    "create", params={"body": params}
                )
                log.info(f'CloudTask Queue [{resource["id"]}] deployed')

            except HttpError as e:
                if e.resp.status == 409:
                    log.info(f'found cloudtask queue {resource["id"]}')

                    if self.should_patch(resource["id"]):
                        self.versioned_clients.cloudtask_queue.execute(
                            "patch",
                            parent_key="name",
                            parent_schema=resource["name"],
                            params={"body": params},
                        )
                        log.info(f"CloudTask Queue [{resource['id']}] patched")
                else:
                    raise e

    def destroy(self):
        if not self.resources:
            return
        for resource_id, resource in self.resources.items():
            try:
                resp = self.versioned_clients.cloudtask_queue.execute(
                    "delete", parent_key="name", parent_schema=resource["name"]
                )
                if resp == {}:
                    log.info(f"CloudTask Queue [{resource['id']}] destroyed")
            except HttpError as e:
                if e.resp.status == 404:
                    log.info(f"cloudtask queue {resource['id']} already destroyed")
                else:
                    raise e

    def should_patch(self, resource_id):
        # if there is user config, there is nothing to compare with
        if not self.resources[resource_id].get("config"):
            return False

        deployed_config = self.get_deployed_config(resource_id)
        for section, configurations in self.resources[resource_id]["config"].items():
            for k, v in configurations.items():
                try:
                    # a value set in the desired config is
                    # different from the value deployed
                    if deployed_config[section][k] != v:
                        return True
                except KeyError:
                    log.info(f"config {section}.{k} not found in deployed config")
                except Exception as e:
                    raise e

        return False

    def get_deployed_config(self, resource_id):
        deployed_config = self.get(resource_id)
        return {
            "rateLimits": deployed_config["rateLimits"],
            "retryConfig": deployed_config["retryConfig"],
        }

    def get(self, resource_id):
        if not self.resources or not resource_id:
            return
        return self.versioned_clients.cloudtask_queue.execute(
            "get", parent_key="name", parent_schema=self.resources[resource_id]["name"]
        )

    def get_config(self, config={}):
        if not self.resources:
            return

        return {"resource_type": self.resource_type, "values": {}}



from __future__ import annotations
import os
import logging
from googleapiclient.errors import HttpError

from goblet.client import VersionedClients

from goblet.backends.cloudfunctionv1 import CloudFunctionV1
from goblet.backends.cloudfunctionv2 import CloudFunctionV2
from goblet.backends.cloudrun import CloudRun

from goblet.handlers.bq_remote_function import BigQueryRemoteFunction
from goblet.handlers.eventarc import EventArc
from goblet.handlers.pubsub import PubSub
from goblet.handlers.routes import Routes
from goblet.handlers.scheduler import Scheduler
from goblet.handlers.cloudtasktarget import CloudTaskTarget
from goblet.handlers.storage import Storage
from goblet.handlers.http import HTTP
from goblet.handlers.jobs import Jobs
from goblet.handlers.uptime import Uptime

from goblet.infrastructures.redis import Redis
from goblet.infrastructures.vpcconnector import VPCConnector
from goblet.infrastructures.alerts import Alerts
from goblet.infrastructures.apigateway import ApiGateway
from goblet.infrastructures.cloudtask import CloudTaskQueue
from goblet.infrastructures.pubsub import PubSubTopic
from goblet.infrastructures.bq_spark_stored_procedure import (
    BigQuerySparkStoredProcedure,
)

from goblet.response import default_missing_route

import goblet.globals as g

from goblet.common_cloud_actions import deploy_custom_role, deploy_service_account

from typing import List

log = logging.getLogger(__name__)
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))

EVENT_TYPES = [
    "all",
    "http",
    "schedule",
    "pubsub",
    "storage",
    "route",
    "eventarc",
    "job",
    "bqremotefunction",
    "cloudtasktarget",
    "uptime",
]

SUPPORTED_BACKENDS = {
    "cloudfunction": CloudFunctionV1,
    "cloudfunctionv2": CloudFunctionV2,
    "cloudrun": CloudRun,
}

SUPPORTED_INFRASTRUCTURES = {
    "redis": Redis,
    "vpcconnector": VPCConnector,
    "cloudtaskqueue": CloudTaskQueue,
    "pubsub_topic": PubSubTopic,
    "bqsparkstoredprocedure": BigQuerySparkStoredProcedure,
}


class Resource_Manager:
    """Core Goblet logic. App entrypoint is the __call__ function which routes the request to the corresonding handler class"""

    def __init__(
        self,
        function_name,
        backend,
        cors=None,
        routes_type="apigateway",
    ):
        self.app_list = []

        self.handlers = {
            "cloudtasktarget": CloudTaskTarget(function_name, backend=backend),
            "route": Routes(
                function_name,
                cors=cors,
                backend=backend,
                routes_type=routes_type,
            ),
            "pubsub": PubSub(function_name, backend=backend),
            "storage": Storage(function_name, backend=backend),
            "eventarc": EventArc(function_name, backend=backend),
            "http": HTTP(function_name, backend=backend),
            "jobs": Jobs(function_name, backend=backend),
            "schedule": Scheduler(function_name, backend=backend),
            "bqremotefunction": BigQueryRemoteFunction(function_name, backend=backend),
            "uptime": Uptime(function_name, backend=backend, routes_type=routes_type),
        }

        self.infrastructure = {
            "cloudtaskqueue": CloudTaskQueue(
                function_name,
                backend=backend,
            ),
            "redis": Redis(
                function_name,
                backend=backend,
            ),
            "vpcconnector": VPCConnector(
                function_name,
                backend=backend,
            ),
            "alerts": Alerts(function_name, backend=backend),
            "apigateway": ApiGateway(function_name, backend=backend),
            "pubsub_topic": PubSubTopic(function_name, backend=backend),
            "bqsparkstoredprocedure": BigQuerySparkStoredProcedure(
                function_name, backend=backend
            ),
        }

        self.middleware_handlers = {
            "before": {},
            "after": {},
        }

        self.error_handlers = {"GobletRouteNotFoundError": default_missing_route}

        self.current_request = None
        self.function_name = function_name

    def __call__(self, request, context=None):
        """Goblet entrypoint"""
        self.current_request = request
        self.request_context = context

        # set var's for added apps
        for added_app in self.app_list:
            added_app.current_request = request
            added_app.request_context = context

        event_type = self.get_event_type(request, context)

        try:
            # call before request middleware
            request = self._call_middleware(
                request, event_type, before_or_after="before"
            )
            response = None
            if event_type not in EVENT_TYPES:
                raise ValueError(f"{event_type} not a valid event type")
            if event_type == "job":
                response = self.handlers["jobs"](request, context)
            if event_type == "schedule":
                response = self.handlers["schedule"](request)
            if event_type == "pubsub":
                response = self.handlers["pubsub"](request, context)
            if event_type == "storage":
                # Storage trigger can be made with @eventarc decorator
                try:
                    response = self.handlers["storage"](request, context)
                except ValueError:
                    event_type = "eventarc"
            if event_type == "route":
                response = self.handlers["route"](request)
            if event_type == "http":
                response = self.handlers["http"](request)
            if event_type == "eventarc":
                response = self.handlers["eventarc"](request)
            if event_type == "bqremotefunction":
                response = self.handlers["bqremotefunction"](request)
            if event_type == "cloudtasktarget":
                response = self.handlers["cloudtasktarget"](request)
            if event_type == "uptime":
                response = self.handlers["uptime"](request)

            # call after request middleware
            response = self._call_middleware(
                response, event_type, before_or_after="after"
            )

        except Exception as e:
            if self.error_handlers.get(e.__class__.__name__):
                return self.error_handlers[e.__class__.__name__](e)
            raise e

        return response

    def __add__(self, other):
        self.app_list.append(other)
        for handler in self.handlers:
            self.handlers[handler] += other.handlers[handler]
        return self

    def combine(self, other):
        return self + other

    def get_event_type(self, request, context=None):
        """Parse event type from the event request and context"""
        if os.environ.get("CLOUD_RUN_TASK_INDEX"):
            return "job"
        if context and context.event_type:
            return context.event_type.split(".")[1].split("/")[0]
        if request.headers.get("X-Goblet-Type") == "schedule":
            return "schedule"
        if request.headers.get("X-Goblet-Uptime-Name"):
            return "uptime"
        if request.headers.get("User-Agent") == "Google-Cloud-Tasks":
            return "cloudtasktarget"
        if request.headers.get("Ce-Type") and request.headers.get("Ce-Source"):
            return "eventarc"
        if (
            request.is_json
            and isinstance(request.json, dict)
            and request.json.get("userDefinedContext")
            and request.json["userDefinedContext"].get("X-Goblet-Name")
        ):
            return "bqremotefunction"
        if (
            request.is_json
            and request.get_json(silent=True)
            and isinstance(request.json, dict)
            and request.json.get("subscription")
            and request.json.get("message")
        ):
            return "pubsub"
        if (
            request.path
            and request.path == "/"
            and not request.headers.get("X-Envoy-Original-Path")
        ):
            return "http"
        if request.path:
            return "route"
        return None

    def _call_middleware(self, event, event_type, before_or_after="before"):
        middleware = self.middleware_handlers[before_or_after].get("all", [])
        middleware.extend(self.middleware_handlers[before_or_after].get(event_type, []))
        for m in middleware:
            event = m(event)

        return event

    def get_infrastructure_config(self):
        configs = []
        for _, v in self.infrastructure.items():
            config = v.get_config()
            if config:
                configs.append(config)
        return configs

    def get_registered_handler_resource_types(self):
        """Get resource types for all deployable registered handlers, http is only non deployable resource"""
        resource_types = []
        for _, v in self.handlers.items():
            if v.resource_type != "http" and v.resources:
                resource_types.append(resource_types)
        return resource_types

    def deploy_handlers(self, source, handlers: List[str] = None):
        """Call each handlers deploy method"""
        if handlers:
            for handler in handlers:
                log.info(f"deploying {handler}")
                self.handlers.get(handler).deploy(
                    source, entrypoint="goblet_entrypoint"
                )
        else:
            for k, v in self.handlers.items():
                log.info(f"deploying {k}")
                v.deploy(source, entrypoint="goblet_entrypoint")

    def destroy_handlers(self, handlers: List[str] = None):
        """Call each handlers destroy method

        Args:
            handlers (List[str], optional):
                List of handler resources, if supplied will only target specified resource.
        """
        if handlers:
            for handler in handlers:
                log.info(f"destroying {handler}")
                self.handlers.get(handler).destroy()
        else:
            for k, v in self.handlers.items():
                log.info(f"destroying {k}")
                v.destroy()

    def sync_handlers(self, dryrun=False, handlers: List[str] = None):
        if handlers:
            for handler in handlers:
                try:
                    self.handlers.get(handler).sync(dryrun)
                except HttpError as e:
                    if e.resp.status == 403:
                        continue
                    raise e
        else:
            for _, v in self.handlers.items():
                try:
                    v.sync(dryrun)
                except HttpError as e:
                    if e.resp.status == 403:
                        continue
                    raise e

    def deploy_infrastructure(self, infras: List[str] = None):
        """Call deploy for each infrastructure"""
        if infras:
            for infra in infras:
                log.info(f"deploying {infra}")
                self.infrastructure.get(infra).deploy()
        else:
            for k, v in self.infrastructure.items():
                log.info(f"deploying {k}")
                v.deploy()

    def destroy_infrastructure(self, infras: List[str] = None):
        """Call destroy method for each infrastructure resource

        Args:
            infras (List[str], optional):
                List of infrastructure resources, if supplied will only target specified resource.
        """
        if infras:
            for infra in infras:
                log.info(f"destroying {infra}")
                self.infrastructure.get(infra).destroy()
        else:
            for k, v in self.infrastructure.items():
                log.info(f"destroying {k}")
                v.destroy()

    def sync_infrastructure(self, dryrun=False, infras: List[str] = None):
        if infras:
            for infra in infras:
                try:
                    self.infrastructure.get(infra).sync(dryrun)
                except HttpError as e:
                    if e.resp.status == 403:
                        continue
                    raise e
        else:
            for _, v in self.infrastructure.items():
                try:
                    v.sync(dryrun)
                except HttpError as e:
                    if e.resp.status == 403:
                        continue
                    raise e

    def is_http(self):
        """Is http determines if additional cloudfunctions will be needed since triggers other than http will require their own
        function"""
        # TODO: move to handlers
        if (
            len(self.handlers["route"].resources) > 0
            or len(self.handlers["schedule"].resources) > 0
            or self.handlers["http"].resources
            or self.handlers["pubsub"].is_http()
            or len(self.handlers["bqremotefunction"].resources) > 0
            or len(self.handlers["cloudtasktarget"].resources) > 0
            or len(self.handlers["uptime"].resources) > 0
        ):
            return True
        return False

    def get_backend_and_check_versions(self, backend: str):
        client_versions = VersionedClients().client_versions
        try:
            backend_class = SUPPORTED_BACKENDS[backend]
        except KeyError:
            raise KeyError(f"Backend {backend} not in supported backends")

        version_key = (
            "cloudfunctions" if backend.startswith("cloudfunction") else backend
        )
        # User selected version
        specified_version = None
        if (
            g.config
            and g.config.client_versions
            and isinstance(g.config.client_versions, dict)
        ):
            specified_version = g.config.client_versions.get(version_key)
        if specified_version:
            if specified_version not in backend_class.supported_versions:
                raise ValueError(
                    f"{version_key} version {client_versions[version_key]} "
                    f"not supported. Valid version(s): {', '.join(backend_class.supported_versions)}."
                )
        else:
            # if not set, set to last in list of supported versions (most recent)
            client_versions[version_key] = backend_class.supported_versions[-1]

        return backend_class

    def create_service_account(self, role):
        iam_role_client = VersionedClients().iam_roles
        deploy_custom_role(iam_role_client, role)
        deploy_service_account(VersionedClients(), self.function_name, role["roleId"])



import json
import logging
from base64 import b64encode
from googleapiclient.errors import HttpError
from goblet.infrastructures.infrastructure import Infrastructure
from goblet.client import VersionedClients
from goblet.permissions import gcp_generic_resource_permissions, add_binding
from goblet.client import get_default_project_number
from goblet.common_cloud_actions import (
    create_pubsub_subscription,
    destroy_pubsub_subscription,
)

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.INFO)


class PubSubClient:
    def __init__(self, topic):
        self.topic = topic

    def publish(self, message):
        message = json.dumps(message)
        return VersionedClients().pubsub_topic.execute(
            "publish",
            parent_key="topic",
            parent_schema=self.topic,
            params={
                "body": {
                    "messages": [{"data": b64encode(message.encode()).decode("utf8")}]
                }
            },
        )


class PubSubTopic(Infrastructure):
    resource_type = "pubsub_topic"
    required_apis = ["pubsub"]
    permissions = gcp_generic_resource_permissions("pubsub", "topics")
    supports_local = True

    def register(self, name, kwargs):
        resource_id = name
        topic_config = kwargs.get("config", None)
        dlq = kwargs.get("dlq", False)
        dlq_pull_subscription = kwargs.get("dlq_pull_subscription", None)

        self.resources[resource_id] = {
            "id": resource_id,
            "name": f"{self.versioned_clients.pubsub_topic.parent}/topics/{name}",
            "config": topic_config,
            "dlq": dlq,
            "dlq_pull_subscription": dlq_pull_subscription,
        }
        return PubSubClient(topic=self.resources[resource_id]["name"])

    def _deploy(self):
        if not self.resources:
            return

        for resource_id, resource in self.resources.items():
            params = {"name": resource["name"]}
            if resource["config"]:
                params = {**params, **resource["config"]}

            try:
                self.versioned_clients.pubsub_topic.execute(
                    "create",
                    parent_key="name",
                    parent_schema=resource["name"],
                    params={"body": params},
                )
                log.info(f'PubSub Topic [{resource["id"]}] deployed')

                if resource["dlq"] is True:
                    # Add IAM roles to use dead-letter topics
                    try:
                        default_pubsub_service_account = f"serviceAccount:service-{get_default_project_number()}@gcp-sa-pubsub.iam.gserviceaccount.com"
                        add_binding(
                            self.versioned_clients.pubsub_topic,
                            resource["name"],
                            "roles/pubsub.publisher",
                            default_pubsub_service_account,
                        )
                    except HttpError as e:
                        if e.resp.status == 403:
                            log.info(
                                f"User is not authorized to add IAM role 'roles/pubsub.publisher' to topic '{resource['name']}' you need to handle this manually."
                            )

                    # Create Pull Subscription for DLQ so messages don't get lost
                    dlq_pull_subscription = resource["dlq_pull_subscription"]
                    log.info(
                        f"Creating pull subscription {dlq_pull_subscription['name']} for DLQ {resource['id']}"
                    )
                    create_pubsub_subscription(
                        client=self.versioned_clients.pubsub,
                        sub_name=dlq_pull_subscription["name"],
                        force_update=False,
                        req_body={
                            "name": dlq_pull_subscription["name"],
                            "topic": resource["name"],
                            **dlq_pull_subscription.get("config", {}),
                        },
                    )

            except HttpError as e:
                if e.resp.status == 409:
                    log.info(f'found pubsub topic {resource["id"]}')
                    updateMask = ",".join(self.paths_to_update(resource_id))
                    if updateMask != "":
                        log.info(f"params: {params}")
                        self.versioned_clients.pubsub_topic.execute(
                            "patch",
                            parent=None,
                            params={
                                "name": self.resources[resource_id]["name"],
                                "body": {"topic": params, "updateMask": updateMask},
                            },
                        )
                        log.info(f"PubSub Topic [{resource['id']}] patched")
                else:
                    raise e

    def destroy(self, config={}):
        if not self.resources:
            return
        for resource_id, resource in self.resources.items():
            try:
                if resource["dlq"] is True:
                    # Delete Pull Subscription for DLQ
                    dlq_pull_subscription = resource["dlq_pull_subscription"]
                    log.info(
                        f"Deleting pull subscription {dlq_pull_subscription['name']} for DLQ {resource['id']}"
                    )
                    destroy_pubsub_subscription(
                        self.versioned_clients.pubsub, dlq_pull_subscription["name"]
                    )

                resp = self.versioned_clients.pubsub_topic.execute(
                    "delete", parent_key="topic", parent_schema=resource["name"]
                )
                if resp == {}:
                    log.info(f"PubSub Topic [{resource['id']}] destroyed")
            except HttpError as e:
                if e.resp.status == 404:
                    log.info(f"PubSub Topic {resource['id']} already destroyed")
                else:
                    raise e

    def paths_to_update(self, resource_id):
        paths = []

        # if there is user config, there is nothing to compare with
        if not self.resources[resource_id].get("config"):
            return paths

        deployed_config = self.get(resource_id)
        for k, v in self.resources[resource_id]["config"].items():
            try:
                # a value set in the desired config is
                # different from the value deployed
                if deployed_config[k] != v:
                    paths.append(k)
            except KeyError:
                paths.append(k)
            except Exception as e:
                raise e

        return paths

    def get(self, resource_id):
        if not self.resources or not resource_id:
            return
        return self.versioned_clients.pubsub_topic.execute(
            "get", parent=None, params={"topic": self.resources[resource_id]["name"]}
        )

    def get_config(self, config={}):
        if not self.resources:
            return

        return {"resource_type": self.resource_type, "values": {}}



import json
from goblet.utils import get_g_dir, nested_update
import os
import logging

log = logging.getLogger("goblet.config")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class GConfig(dict):
    """Config class used to get variables from config.json or from the environment. If stage is set as an environment level
    if will parse the corresponding section in config.json and return those config values
    """

    def __init__(self, config=None, stage=None, init=True):
        if config:
            dict.__init__(self, **config)

        if not init:
            self.config = config
            return

        self.config = self.get_g_config()
        if config:
            self.config = nested_update(self.config, config)
        self.stage = stage or os.environ.get("STAGE")
        self.update_stage_config()
        self.validate()

    def update_g_config(self, stage=None, values={}, write_config=False):
        self.stage = self.stage or stage
        if write_config:
            if self.stage:
                self.config["stages"][self.stage] = nested_update(
                    self.config.get("stages", {}).get(self.stage, {}), values
                )
            else:
                self.config = nested_update(self.config, values)
            self.write()
        self.update_stage_config()
        self.config = nested_update(self.config, values)

    def update_stage_config(self):
        if self.stage:
            self.config = nested_update(
                self.config, self.config.get("stages", {}).get(self.stage, {})
            )

    @staticmethod
    def get_g_config():
        try:
            with open(f"{get_g_dir()}/config.json") as f:
                return json.load(f)
        except FileNotFoundError:
            return {}
        except json.decoder.JSONDecodeError:
            log.info(
                "JSONDecodeError. config.json is not valid. Returning empty config"
            )
            return {}

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        self.config[key] = value

    def __getattr__(self, name):
        if os.environ.get(name):
            return os.environ.get(name)
        attr = self.config.get(name)
        if isinstance(attr, dict) and len(attr) > 0:
            return GConfig(config=attr, init=False)
        if attr is None:
            return GConfig(config={}, init=False)
        if attr or attr == {}:
            return attr
        return None

    def __setattr__(self, name, value):
        if name not in ["config", "stage"]:
            self.config[name] = value
        else:
            super(GConfig, self).__setattr__(name, value)

    def write(self):
        with open(f"{get_g_dir()}/config.json", "w") as f:
            f.write(json.dumps(self.config, indent=4))

    def validate(self):
        if self.stage and self.stage not in self.config.get("stages", {}):
            raise ValueError(f"stage {self.stage} not found in config")



import logging
import os

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


def gcp_generic_resource_permissions(service, subservice):
    """
    Basic grouping of resource permissions
    """
    return [
        f"{service}.{subservice}.create",
        f"{service}.{subservice}.get",
        f"{service}.{subservice}.delete",
        f"{service}.{subservice}.update",
        f"{service}.{subservice}.list",
    ]


def create_custom_role_policy(app_name, permissions):
    """
    https://cloud.google.com/iam/docs/creating-custom-roles#iam-custom-roles-create-rest
    Role name much be `"[a-zA-Z0-9_.]{3,64}"`
    """
    return {
        "roleId": f"Goblet_Deployment_Role_{app_name.replace('-','_')}",
        "role": {
            "title": f"Deployment role for {app_name}",
            "description": "Goblet generated role",
            "includedPermissions": permissions,
            "stage": "GA",
        },
    }


def add_binding(client, resource_parent_schema, roleName, principals):
    """Generic add-binding procedure that updates the current resource policy if the desired role and principle do not already exist."""
    # Get Iam policy for resource
    resp = client.execute(
        "getIamPolicy", parent_key="resource", parent_schema=resource_parent_schema
    )
    bindings = resp.get("bindings", [])
    # default service account type for bindings
    principals = [f"serviceAccount:{p}" if ":" not in p else p for p in principals]
    role_missing = True
    # Check to see if desired role and principle exist
    for role_binding in bindings:
        if role_binding["role"] == roleName:
            if all(p in role_binding["members"] for p in principals):
                # Member exists
                log.info(
                    f"iam policy for {resource_parent_schema} already up to date..."
                )
                return
            else:
                role_missing = False
                role_binding["members"].extend(principals)
    if role_missing:
        bindings.append({"role": roleName, "members": principals})
    log.info(f"setting iam policy for {resource_parent_schema}...")
    client.execute(
        "setIamPolicy",
        parent_key="resource",
        parent_schema=resource_parent_schema,
        params={"body": {"policy": {"bindings": bindings}}},
    )



import os
import re
import base64
from functools import lru_cache
from urllib.parse import quote_plus

import google_auth_httplib2
from googleapiclient.errors import HttpError

from goblet.backends.backend import Backend
from goblet.client import VersionedClients
from goblet_gcp_client.client import (
    get_default_project,
    get_default_location,
    Client,
    get_credentials,
)
from goblet.common_cloud_actions import (
    create_cloudbuild,
    destroy_cloudrun,
    destroy_cloudfunction_artifacts,
    get_cloudrun_url,
    getDefaultRegistry,
    getDefaultRegistryName,
)
from goblet.revision import RevisionSpec
from goblet.utils import get_dir
from goblet.write_files import write_dockerfile
from goblet.errors import GobletValidationError
from goblet.permissions import gcp_generic_resource_permissions, add_binding


class CloudRun(Backend):
    resource_type = "cloudrun"
    supported_versions = ["v2"]
    monitoring_type = "cloud_run_revision"
    monitoring_label_key = "service_name"
    required_apis = ["run", "cloudbuild", "cloudfunctions", "cloudresourcemanager"]
    permissions = [
        *gcp_generic_resource_permissions("run", "services"),
        "run.services.getIamPolicy",
        "run.services.setIamPolicy",
        "run.revisions.get",
        "run.revisions.list",
        "run.operations.get",
        "cloudbuild.builds.create",
        "cloudbuild.builds.get",
        "cloudbuild.builds.list",
        "cloudresourcemanager.projects.get",
        "iam.serviceaccounts.actAs",
        "cloudfunctions.functions.sourceCodeSet",
        "artifactregistry.repositories.create",
        "artifactregistry.repositories.get",
    ]

    def __init__(self, app):
        self.client = VersionedClients().run
        self.run_name = f"projects/{get_default_project()}/locations/{get_default_location()}/services/{app.function_name}"
        super().__init__(app, self.client, self.run_name)

    def validation_config(self):
        name_pattern = r"^[a-z]([-a-z0-9]*[a-z0-9])?"
        pattern = re.compile(name_pattern)
        if not re.fullmatch(pattern, self.name):
            raise GobletValidationError(
                f"Invalid Cloudrun name {self.name}. Needs to follow regex of pattern {name_pattern}"
            )

    def deploy(self, force=False):
        versioned_clients = VersionedClients()
        put_headers = {
            "content-type": "application/zip",
        }

        if not os.path.exists(
            get_dir() + f"/{self.config.dockerfile or 'Dockerfile'}"
        ) and not os.path.exists(get_dir() + "/Procfile"):
            self.log.info(
                "No Dockerfile or Procfile found for cloudrun backend. Writing default Dockerfile"
            )
            write_dockerfile()

        try:
            artifact_tag = os.environ["GOBLET_ARTIFACT_TAG"]
        except KeyError:
            artifact_tag = self.config.deploy.get("artifact_tag", None)

        if artifact_tag:
            source, changes = None, False
            self.log.info(
                f"skipping zip/upload/build... cloudbuild.artifact {artifact_tag} found"
            )
        else:
            self._zip_file(self.config.dockerfile or "Dockerfile", "Dockerfile")
            source, changes = self._gcs_upload(
                self.client,
                put_headers,
                upload_client=versioned_clients.run_uploader,
                force=force,
            )

            if not changes:
                return None

            self.create_build(versioned_clients.cloudbuild, source, self.name)

        if not self.skip_run_deployment():
            serviceRevision = RevisionSpec(self.config, versioned_clients, self.name)
            serviceRevision.deployRevision()
        else:
            self.log.info("Skipping cloudrun deployment since it is not needed...")

        # Set IAM Bindings
        if not self.skip_run_deployment() and self.config.bindings:
            self.log.info(f"adding IAM bindings for cloudrun {self.name}")
            policy_bindings = {"policy": {"bindings": self.config.bindings}}
            self.client.execute(
                "setIamPolicy",
                parent_key="resource",
                parent_schema=self.run_name,
                params={"body": policy_bindings},
            )

        return source

    def destroy(self, all=False):
        destroy_cloudrun(self.client, self.name)
        if all:
            destroy_cloudfunction_artifacts(self.name)

    def create_build(self, client, source=None, name="goblet"):
        """Creates http cloudbuild"""
        build_configs = self.config.cloudbuild.copy() if self.config.cloudbuild else {}

        try:
            registry = self.config.deploy.get(
                "artifact_registry"
            ) or getDefaultRegistry(name)
        except AttributeError:
            registry = getDefaultRegistry(name)

        # check if default registry exists
        if registry == getDefaultRegistry(name):
            registry_client = VersionedClients().artifactregistry_repositories
            try:
                registry_client.execute(
                    "get", parent_key="name", parent_schema=getDefaultRegistryName()
                )
            except HttpError as e:
                # Registry doesn't exist
                if e.resp.status == 404:
                    # create registry
                    self.log.info(
                        f"Default registry doesn't exist. Creating registry {getDefaultRegistryName()}"
                    )
                    resp = registry_client.execute(
                        "create",
                        params={
                            "body": {
                                "name": getDefaultRegistryName(),
                                "format": "DOCKER",
                                "mode": "STANDARD_REPOSITORY",
                            },
                            "repositoryId": "cloud-run-source-deploy",
                        },
                    )
                registry_client.wait_for_operation(resp["name"])

        if build_configs.get("serviceAccount") and not build_configs.get("logsBucket"):
            build_options = build_configs.get("options", {})
            if not build_options.get("logging"):
                build_options["logging"] = "CLOUD_LOGGING_ONLY"
                build_configs["options"] = build_options
                self.log.info(
                    "service account given but no logging bucket so defaulting to cloud logging only"
                )

        images = [f"{registry}:latest"]
        # Add environment variable tags
        build_tags = os.environ.get("GOBLET_BUILD_TAGS", None)
        if build_tags:
            images += list(map(lambda tag: f"{registry}:{tag}", build_tags.split(",")))

        steps = self._get_cloudbuild_steps(images)

        req_body = {
            "source": {"storageSource": source["storageSource"]},
            "steps": steps,
            "images": images,
            **build_configs,
        }

        req_body["tags"] = build_configs.get("tags", []) + [f"goblet-build-{self.name}"]

        create_cloudbuild(client, req_body)

    def skip_deployment(self):
        return self.skip_run_deployment

    def skip_run_deployment(self):
        """Skip cloudrun deployment if only jobs"""
        skip = True
        for name, handler in self.app.handlers.items():
            if handler.resources and name != "jobs" and name != "schedule":
                skip = False
            # scheduled jobs
            if handler.resources and name == "schedule":
                for schedule_name in handler.resources.keys():
                    if not schedule_name.startswith("schedule-job-"):
                        skip = False
        # Forces the deployment of cloud run
        if self.config.force_deploy_cloudrun:
            return False
        return skip

    def _checksum(self):
        versioned_clients = VersionedClients()
        resp = versioned_clients.cloudbuild.execute(
            "list",
            parent_key="projectId",
            parent_schema=get_default_project(),
            params={"filter": f"tags=goblet-build-{self.name}"},
        )
        if not resp:
            return 0
        latest_build_source = resp["builds"][0].get("source")
        if not latest_build_source:
            return 0
        bucket = latest_build_source["storageSource"]["bucket"]
        obj = latest_build_source["storageSource"]["object"]
        client = Client("cloudresourcemanager", "v1", calls="projects")
        http = client.http or google_auth_httplib2.AuthorizedHttp(get_credentials())
        resp = http.request(
            f"https://storage.googleapis.com/storage/v1/b/{bucket}/o/{quote_plus(obj)}?alt=media",
        )
        try:
            return resp[0]["x-goog-hash"].split(",")[-1].split("=", 1)[-1]
        except KeyError:
            return 0

    def _get_cloudbuild_steps(self, images):
        cloudbuild_cache = self.config.deploy.get("cloudbuild_cache", "DOCKER_LATEST")

        build_args = os.environ.get("GOBLET_BUILD_ARGS", [])
        if build_args:
            build_args = build_args.split(",")

        if cloudbuild_cache == "DOCKER_LATEST":
            steps = [
                {
                    "name": "gcr.io/cloud-builders/docker",
                    "entrypoint": "bash",
                    "args": ["-c", f"docker pull {images[0]} || exit 0"],
                },
                {
                    "name": "gcr.io/cloud-builders/docker",
                    "args": [
                        "build",
                        "--network=cloudbuild",
                    ]
                    + list(map(lambda image: ["-t", image], images))
                    + list(
                        map(lambda build_arg: ["--build-arg", build_arg], build_args)
                    )
                    + [
                        "--cache-from",
                        images[0],
                        ".",
                    ],
                },
            ]
        elif cloudbuild_cache == "KANIKO":
            steps = [
                {
                    "name": "gcr.io/kaniko-project/executor:latest",
                    "args": [
                        f"--destination={images[0]}",
                        "--cache=true",
                        f"--cache-ttl={24*365*3}h",
                    ]
                    + [["--build-arg", f"{build_arg}"] for build_arg in build_args],
                },
                {
                    "name": "gcr.io/cloud-builders/docker",
                    "entrypoint": "bash",
                    "args": [
                        "-c",
                        f"docker pull {images[0]}"
                        + " && ".join(
                            [""]
                            + [
                                f"docker tag {images[0]} {build_tag_image}"
                                for build_tag_image in images[1:]
                            ]
                        ),
                    ],
                },
            ]

        else:
            raise Exception(f"Invalid cloudbuild_cache option {cloudbuild_cache}")

        return steps

    def update_config(self, infra_configs=[], write_config=False, stage=None):
        config_updates = {}
        for infra_config in infra_configs:
            resource_type = infra_config["resource_type"]
            if resource_type == "vpcconnector":
                config_updates["cloudrun_revision"] = {
                    **config_updates.get("cloudrun_revision", {}),
                    "vpcAccess": {
                        "connector": infra_config["values"]["name"],
                        "egress": infra_config["values"]["egress"],
                    },
                }
            else:
                envs = [
                    {"name": name, "value": value}
                    for name, value in infra_config["values"].items()
                ]
                env = config_updates.get("cloudrun_container", {}).get("env", [])
                env.extend(envs)

                config_updates["cloudrun_container"] = {
                    **config_updates.get("cloudrun_container", {}),
                    "env": env,
                }
        self.config.update_g_config(
            values=config_updates,
            write_config=write_config,
            stage=stage,
        )

    @property
    @lru_cache(maxsize=1)
    def http_endpoint(self):
        return get_cloudrun_url(self.client, self.name)

    def set_iam_policy(self, service_account_id):
        client = self.client
        policy = {
            "policy": {
                "bindings": {
                    "role": "roles/run.invoker",
                    "members": [f"serviceAccount:{service_account_id}"],
                }
            }
        }
        client.execute(
            "setIamPolicy",
            params={"body": policy},
            parent_key="resource",
            parent_schema=self.run_name,
        )

    def get_environment_vars(self):
        env_dict = {}
        env = self.config.config.get("cloudrun_container", {}).get("env", [])
        # Append if job_container is set
        env.extend(self.config.config.get("job_container", {}).get("env", []))

        versioned_clients = VersionedClients()
        for env_item in env:
            # get secret
            if env_item.get("valueSource"):
                try:
                    secret_name = env_item["valueSource"]["secretKeyRef"]["secret"]
                    version = env_item["valueSource"]["secretKeyRef"].get(
                        "version", "latest"
                    )

                    resp = versioned_clients.secretmanager.execute(
                        "access",
                        parent_key="name",
                        parent_schema="projects/{project_id}/secrets/"
                        + secret_name
                        + "/versions/"
                        + version,
                    )

                    env_dict[env_item["name"]] = base64.b64decode(
                        resp["payload"]["data"]
                    ).decode()
                except Exception as e:
                    self.log.info(
                        f"Unable to get secret {env_item['name']} and set environment variable with error {str(e)}. Skipping..."
                    )
            else:
                env_dict[env_item["name"]] = env_item["value"]
        return env_dict

    def zip_required_files(self):
        """Zip required files for cloudrun. Requirements.txt is not required."""
        self._zip_config()
        if self.config.requirements_file:
            self._zip_file(self.config.requirements_file, "requirements.txt")
        if self.config.main_file:
            self._zip_file(self.config.main_file, "main.py")

    def add_invoker_binding(self, principles):
        add_binding(self.client, self.run_name, "roles/run.invoker", principles)



import os
import importlib.util
import collections.abc
from contextlib import contextmanager
import sys
import json
import tempfile


@contextmanager
def add_to_path(p):
    import sys

    old_path = sys.path
    sys.path = sys.path[:]
    sys.path.insert(0, p)
    try:
        yield
    finally:
        sys.path = old_path


def checksum(fh, hasher, blocksize=65536):
    """Calculates checksum of file"""
    buf = fh.read(blocksize)
    while len(buf) > 0:
        hasher.update(buf)
        buf = fh.read(blocksize)
    return hasher.digest()


def get_app_from_module(m):
    from goblet import Goblet

    for obj in dir(m):
        if isinstance(getattr(m, obj), Goblet) and not getattr(m, obj).is_sub_app:
            return getattr(m, obj), obj


def get_goblet_app(main_file="main.py"):
    """Look for main.py or main_file if defined and return goblet app instance."""
    dir_path = os.path.realpath(".")
    spec = importlib.util.spec_from_file_location("main", f"{dir_path}/{main_file}")
    main = importlib.util.module_from_spec(spec)
    with add_to_path(dir_path):
        spec.loader.exec_module(main)
        app, app_name = get_app_from_module(main)
    # setattr(app, "entrypoint", app_name)
    return app


def get_g_dir():
    """Gets the .goblet directory"""
    return f"{os.path.realpath('.')}/.goblet"


def get_dir():
    return os.path.realpath(".")


def build_stage_config(config_path, stage):
    with open(config_path) as f:
        config = json.load(f)
        stage_config = config["stages"][stage]
        del config["stages"]
        config = nested_update(config, stage_config)

    config_file = tempfile.NamedTemporaryFile(mode="w")
    f = open(config_file.name, "w")
    f.write(json.dumps(config, indent=4))
    f.flush()
    return config_file


def nested_update(d, u):
    """
    Updates nested dictionary d with nested dictionary u
    """
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = nested_update(d.get(k, {}), v)
        elif isinstance(v, list):
            d[k] = [] if not d.get(k) else d[k]
            for obj in v:
                if obj not in d[k]:
                    d[k].append(obj)
        else:
            d[k] = v
    return d


def attributes_to_filter(attributes: dict) -> str:
    """
    Returns pubsub filter based on attributes
    """
    attribute_list = [
        f'attributes.{attribute} = "{value}"' for attribute, value in attributes.items()
    ]
    pubsub_filter = " OR ".join(attribute_list)
    return pubsub_filter


def get_python_runtime() -> str:
    """
    Returns pythons runtime in cloudfunction format ex python37
    """
    version_info = sys.version_info
    return f"python{version_info.major}{version_info.minor}"



import json
import os

from goblet.utils import get_g_dir, get_dir
from goblet.__version__ import __version__


def create_goblet_dir(name, backend):
    """Creates a new goblet directory with a sample main.py, requirements.txt, and config.json"""
    try:
        os.mkdir(get_g_dir())
    except FileExistsError:
        pass
    with open(f"{get_g_dir()}/config.json", "w") as f:
        if backend == "cloudrun":
            f.write(json.dumps({"cloudrun": {}}, indent=4))
        else:
            f.write(json.dumps({"cloudfunction": {}}, indent=4))
    with open("requirements.txt", "w") as f:
        f.write(f"goblet-gcp=={__version__}")
    with open("main.py", "w") as f:
        f.write(
            f"""from goblet import Goblet, jsonify,goblet_entrypoint

app = Goblet(function_name="goblet-{name}", backend="{backend}")
goblet_entrypoint(app)

@app.http()
def main(request):
    return jsonify(request.json)

# route
# @app.route('/hello')
# def home():
#     return jsonify("goodbye")

# schedule
# @app.schedule('5 * * * *')
# def scheduled_job():
#     return jsonify("success")

# pubsub topic
# @app.pubsub_subscription('test_topic')
# def topic(data):
#     app.log.info(data)
#     return
"""
        )
    with open("README.md", "w") as f:
        f.write(
            f"""# goblet-{name}

autocreated by goblet

To test endpoints locally run `goblet local`
To deploy cloudfunctions and other gcp resources defined in `main.py` run `goblet deploy`

To check out goblet documentation go to [docs](https://goblet.github.io/goblet/build/html/index.html)
"""
        )


def write_dockerfile():
    with open(f"{get_dir()}/Dockerfile", "w") as f:
        f.write(
            """\
# https://hub.docker.com/_/python
FROM python:3.10-slim

# setup environment
ENV APP_HOME /app
WORKDIR $APP_HOME

# install keyring backend to handle artifact registry authentication
# RUN pip install keyrings.google-artifactregistry-auth==1.1.1

# Install dependencies.
COPY requirements.txt .
RUN pip install -r requirements.txt

# Copy local code to the container image.
COPY . .

# Run the web service on container startup.
CMD exec functions-framework --target=goblet_entrypoint
"""
        )



from goblet.config import GConfig
import click
import os
import logging
import subprocess
import json
import requests
import sys
import yaml as yml

from goblet.utils import get_g_dir, get_goblet_app
from goblet.write_files import create_goblet_dir
from goblet_gcp_client.client import get_default_project
from goblet.__version__ import __version__
import goblet.globals as g
from goblet.permissions import create_custom_role_policy

logging.basicConfig()

SUPPORTED_BACKENDS = ["cloudfunction", "cloudfunctionv2", "cloudrun"]
SUPPORTED_HANDLERS = [
    "bqremotefunction",
    "cloudtasktarget",
    "eventarc",
    "jobs",
    "pubsub",
    "storage",
    "schedule",
    "uptime",
]
SUPPORTED_INFRASTRUCTURES = [
    "alerts",
    "apigateway",
    "cloudtaskqueue",
    "pubsub",
    "redis",
    "vpcconnector",
    "bqsparkstoredprocedure",
]


@click.group()
@click.option("--debug", is_flag=True)
def main(debug):
    if debug:
        click.echo("Debug mode is 'on'\n")
        os.environ["GOBLET_LOG_LEVEL"] = "DEBUG"
    pass


@main.command()
def help():
    click.echo(
        "Use goblet --help. You can also view the full docs for goblet at https://goblet.github.io/goblet/docs/build/html/index.html"
    )


@main.command()
def version():
    """current goblet version"""
    click.echo(__version__)


@main.command()
@click.option("-p", "--project", "project", envvar="GOOGLE_PROJECT")
@click.option("-l", "--location", "location", envvar="GOOGLE_LOCATION", required=True)
@click.option("-s", "--stage", "stage", envvar="STAGE")
@click.option("--skip-handlers", "skip_handlers", is_flag=True)
@click.option("--skip-backend", "skip_backend", is_flag=True)
@click.option("--skip-infra", "skip_infra", is_flag=True)
@click.option("--config-from-json-string", "config")
@click.option("-f", "--force", "force", is_flag=True)
@click.option("--write-config", "write_config", is_flag=True)
@click.option(
    "-h", "--handler", "handler", type=click.Choice(SUPPORTED_HANDLERS), multiple=True
)
@click.option(
    "-i",
    "--infra",
    "infra",
    type=click.Choice(SUPPORTED_INFRASTRUCTURES),
    multiple=True,
)
def deploy(
    project,
    location,
    stage,
    skip_handlers,
    skip_backend,
    skip_infra,
    config,
    force,
    write_config,
    handler,
    infra,
):
    """
    You can set the project and location using environment variable GOOGLE_PROJECT and GOOGLE_LOCATION

    Note: Allowed GOOGLE_LOCATION values for API GATEWAY are: asia-east1, europe-west1, us-eastl1 and us-central1.

    Note: Make sure api-gateway, cloudfunctions, and storage are enabled in your project
    """
    os.environ["X-GOBLET-DEPLOY"] = "true"
    try:
        _project = project or get_default_project()
        if not _project:
            click.echo(
                "Project not found. Set --project flag or add to gcloud by using gcloud config set project PROJECT"
            )
        os.environ["GOOGLE_PROJECT"] = _project
        os.environ["GOOGLE_LOCATION"] = location
        if stage:
            os.environ["STAGE"] = stage

        # import config from string
        imported_config = {}
        if config:
            imported_config = json.loads(config)

        # get goblet config
        goblet_config = GConfig(imported_config)

        # set deploy env vars
        if goblet_config.deploy:
            for key, value in goblet_config.deploy.get(
                "environmentVariables", {}
            ).items():
                os.environ[key] = value

        app = get_goblet_app(goblet_config.main_file or "main.py")

        # update config
        g.config.update_g_config(
            values=imported_config, write_config=write_config, stage=stage
        )

        app.deploy(
            skip_handlers,
            skip_backend,
            skip_infra,
            force=force,
            stage=stage,
            write_config=write_config,
            handlers=list(set(handler)),
            infras=list(set(infra)),
        )

    except FileNotFoundError as not_found:
        click.echo(
            f"Missing {not_found.filename}. Make sure you are in the correct directory and this file exists"
        )
        sys.exit(1)


@main.command()
@click.option("-p", "--project", "project", envvar="GOOGLE_PROJECT")
@click.option("-l", "--location", "location", envvar="GOOGLE_LOCATION", required=True)
@click.option("-s", "--stage", "stage", envvar="STAGE")
@click.option("-a", "--all", "all", is_flag=True)
@click.option("--skip-infra", "skip_infra", is_flag=True)
@click.option("--skip-handlers", "skip_handlers", is_flag=True)
@click.option("--skip-backend", "skip_backend", is_flag=True)
@click.option(
    "-h", "--handler", "handler", type=click.Choice(SUPPORTED_HANDLERS), multiple=True
)
@click.option(
    "-i",
    "--infra",
    "infra",
    type=click.Choice(SUPPORTED_INFRASTRUCTURES),
    multiple=True,
)
def destroy(
    project,
    location,
    stage,
    all,
    skip_infra,
    skip_handlers,
    skip_backend,
    handler,
    infra,
):
    """
    Deletes all resources in gcp that are defined the current deployment

    The --all flag removes cloudfunction artifacts in cloud storage as well
    """
    os.environ["X-GOBLET-DEPLOY"] = "true"
    try:
        _project = project or get_default_project()
        if not _project:
            click.echo(
                "Project not found. Set --project flag or add to gcloud by using gcloud config set project PROJECT"
            )
        os.environ["GOOGLE_PROJECT"] = _project
        os.environ["GOOGLE_LOCATION"] = location
        if stage:
            os.environ["STAGE"] = stage

        # get goblet config
        goblet_config = GConfig()

        # set deploy env vars
        if goblet_config.deploy:
            for key, value in goblet_config.deploy.get(
                "environmentVariables", {}
            ).items():
                os.environ[key] = value

        app = get_goblet_app(goblet_config.main_file or "main.py")
        app.destroy(
            all,
            skip_infra,
            skip_handlers,
            skip_backend,
            handlers=list(set(handler)),
            infras=list(set(infra)),
        )

    except FileNotFoundError as not_found:
        click.echo(
            f"Missing {not_found.filename}. Make sure you are in the correct directory and this file exists"
        )
        sys.exit(1)


@main.command()
@click.option("-p", "--project", "project", envvar="GOOGLE_PROJECT")
@click.option("-l", "--location", "location", envvar="GOOGLE_LOCATION", required=True)
@click.option("-s", "--stage", "stage", envvar="STAGE")
@click.option("-d", "--dryrun", "dryrun", is_flag=True)
@click.option("--skip-infra", "skip_infra", is_flag=True)
@click.option("--skip-handlers", "skip_handlers", is_flag=True)
@click.option(
    "-h", "--handler", "handler", type=click.Choice(SUPPORTED_HANDLERS), multiple=True
)
@click.option(
    "-i",
    "--infra",
    "infra",
    type=click.Choice(SUPPORTED_INFRASTRUCTURES),
    multiple=True,
)
def sync(
    project,
    location,
    stage,
    dryrun,
    skip_infra,
    skip_handlers,
    handler,
    infra,
):
    """
    Syncs resources that are deployed with current app configuration. This command will delete resources based on naming
    convention that are no longer in the app configuration.

    Use --dryrun flag to see what resources are flagged as being deleted.
    """
    os.environ["X-GOBLET-DEPLOY"] = "true"
    try:
        _project = project or get_default_project()
        if not _project:
            click.echo(
                "Project not found. Set --project flag or add to gcloud by using gcloud config set project PROJECT"
            )
        os.environ["GOOGLE_PROJECT"] = _project
        os.environ["GOOGLE_LOCATION"] = location
        if stage:
            os.environ["STAGE"] = stage
        app = get_goblet_app(GConfig().main_file or "main.py")
        app.sync(dryrun, skip_infra, skip_handlers, handler, infra)

    except FileNotFoundError as not_found:
        click.echo(
            f"Missing {not_found.filename}. Make sure you are in the correct directory and this file exists"
        )
        sys.exit(1)


@main.command()
@click.argument("cloudfunction")
@click.option("-s", "--stage", "stage", envvar="STAGE")
@click.option("-v", "--version", "version", envvar="VERSION", type=click.Choice("3"))
def openapi(cloudfunction, stage, version):
    """
    You can find the generated openapi spec in /.goblet folder.

    The cloudfunction argument sets the correct x-google-backend address in the openapi spec.
    """
    os.environ["X-GOBLET-DEPLOY"] = "true"
    if stage:
        os.environ["STAGE"] = stage
    try:
        app = get_goblet_app(GConfig().main_file or "main.py")
        app.handlers["route"].generate_openapi_spec(cloudfunction)
    except FileNotFoundError as not_found:
        click.echo(
            f"Missing {not_found.filename}. Make sure you are in the correct directory and this file exists"
        )
        sys.exit(1)

    if version:
        with open(f"{get_g_dir()}/{app.function_name}_openapi_spec.yml", "r") as f:
            data = f.read()
        headers = {
            "accept": "application/yaml",
            "Content-Type": "application/yaml",
        }
        response = requests.post(
            "https://converter.swagger.io/api/convert", headers=headers, data=data
        )
        with open(f"{get_g_dir()}/{app.function_name}_openapi_spec_3.yml", "w+") as f:
            f.write(response.text)


@main.command()
@click.argument("local_arg", default="local")
@click.option("-s", "--stage", "stage", envvar="STAGE")
@click.option("-p", "--port", "port", envvar="PORT", default=8080)
@click.option("--set-env", "set_env", is_flag=True)
@click.option("--extras", "extras", is_flag=True)
def local(local_arg, stage, port, set_env, extras):
    """
    Requires the local argument to be set in the Goblet class. The default is local.

    For example in this case you would use local_function

    Goblet("test_function",local="local_function")
    """
    os.environ["X_GOBLET_LOCAL"] = "true"
    try:
        if stage:
            os.environ["STAGE"] = stage
        config = GConfig()
        source = config.main_file or "main.py"
        if set_env:
            app = get_goblet_app(source)
            env_dict = app.backend.get_environment_vars()
            for k, v in env_dict.items():
                os.environ[k] = v
        if extras:
            app = get_goblet_app(source)
            app.deploy_local()
        subprocess.check_output(
            [
                "functions-framework",
                f"--target={local_arg}",
                "--debug",
                f"--source={source}",
                f"--port={port}",
            ]
        )
    except subprocess.CalledProcessError:
        click.echo(
            "Incorrect argument. Make sure you set the local param in your Goblet class and that it matches the arg used in goblet local"
        )


@click.option("-s", "--stage", "stage", envvar="STAGE")
@main.command()
def package(stage):
    """generates the goblet zipped package in .goblet folder"""
    os.environ["X-GOBLET-DEPLOY"] = "true"
    try:
        if stage:
            os.environ["STAGE"] = stage
        app = get_goblet_app(GConfig().main_file or "main.py")
        app.package()
        app.backend.zipf.close()

    except FileNotFoundError as not_found:
        click.echo(
            f"Missing {not_found.filename}. Make sure you are in the correct directory and this file exists"
        )
        sys.exit(1)


@click.argument("name")
@click.option(
    "-b",
    "--backend",
    "backend",
    default="cloudfunction",
    type=click.Choice(SUPPORTED_BACKENDS),
)
@main.command()
def init(name, backend):
    """Create new goblet app with files main.py, requirements.txt, and directory .goblet"""
    create_goblet_dir(name, backend)
    click.echo("created .goblet/config.json")
    click.echo("created requirements.txt")
    click.echo("created main.py")
    click.echo("created README.md")


@main.group()
def stage():
    """view and create different environment stages"""
    pass


@stage.command(name="list")
def list_stages():
    config = GConfig.get_g_config()
    if not config.get("stages"):
        return click.echo("no stages found")
    click.echo(list(config["stages"].keys()))


@stage.command()
@click.argument(
    "stage",
)
def create(stage):
    """create a new stage in config.json"""
    os.environ["X-GOBLET-DEPLOY"] = "true"
    config = GConfig()
    if config.stages and stage in config.stages:
        return click.echo(f"stage {stage} already exists")
    app = get_goblet_app(GConfig().main_file or "main.py")
    function_name = f"{app.function_name}-{stage}"
    if not config.stages:
        config.stages = {stage: {"function_name": function_name}}
    else:
        config.stages[stage] = {"function_name": function_name}
    config.write()
    click.echo(
        f"stage {stage} created in config.json with function name {function_name}"
    )


@main.group()
def job():
    """run cloudrun jobs"""
    pass


@job.command(name="run")
@click.argument(
    "name",
)
@click.argument("task_id", envvar="CLOUD_RUN_TASK_INDEX", default="0")
@click.option("--set-env", "set_env", is_flag=True)
@click.option("-s", "--stage", "stage", envvar="STAGE")
def run_job(name, task_id, set_env, stage):
    """
    Run a Cloudrun Job in local environment.
    """
    os.environ["CLOUD_RUN_TASK_INDEX"] = task_id

    try:
        app = get_goblet_app(GConfig().main_file or "main.py")
        if stage:
            os.environ["STAGE"] = stage
        if set_env:
            env_dict = app.backend.get_environment_vars()
            for k, v in env_dict.items():
                os.environ[k] = v
        app(name, int(task_id))

    except FileNotFoundError as not_found:
        click.echo(
            f"Missing {not_found.filename}. Make sure you are in the correct directory and this file exists"
        )
        sys.exit(1)


@main.group()
def services():
    """check and enable gcp service apis for your gcp project"""
    pass


@services.command(name="check")
@click.option("-p", "--project", "project", envvar="GOOGLE_PROJECT")
@click.option("-s", "--stage", "stage", envvar="STAGE")
def check_gcp_services(project, stage):
    """check status of gcp service apis"""
    os.environ["X-GOBLET-DEPLOY"] = "true"
    try:
        _project = project or get_default_project()
        if not _project:
            click.echo(
                "Project not found. Set --project flag or add to gcloud by using gcloud config set project PROJECT"
            )
        os.environ["GOOGLE_PROJECT"] = _project
        if stage:
            os.environ["STAGE"] = stage
        app = get_goblet_app(GConfig().main_file or "main.py")
        app.check_or_enable_services()
    except FileNotFoundError as not_found:
        click.echo(
            f"Missing {not_found.filename}. Make sure you are in the correct directory and this file exists"
        )
        sys.exit(1)


@services.command(name="enable")
@click.option("-p", "--project", "project", envvar="GOOGLE_PROJECT")
@click.option("-s", "--stage", "stage", envvar="STAGE")
def enable_gcp_services(project, stage):
    """enable gcp service apis"""
    os.environ["X-GOBLET-DEPLOY"] = "true"
    try:
        _project = project or get_default_project()
        if not _project:
            click.echo(
                "Project not found. Set --project flag or add to gcloud by using gcloud config set project PROJECT"
            )
        os.environ["GOOGLE_PROJECT"] = _project
        if stage:
            os.environ["STAGE"] = stage
        app = get_goblet_app(GConfig().main_file or "main.py")
        app.check_or_enable_services(enable=True)

    except FileNotFoundError as not_found:
        click.echo(
            f"Missing {not_found.filename}. Make sure you are in the correct directory and this file exists"
        )
        sys.exit(1)


@services.command(name="autogen_iam")
@click.option("--yaml", "yaml", is_flag=True)
@click.option("-s", "--stage", "stage", envvar="STAGE")
def autogen_iam(yaml, stage):
    """Generate a custom role json or yaml

    Can create the custom role using gcloud cli using the yaml output:

    gcloud iam roles create myCompanyAdmin --organization=123456789012 --file=my-role-definition.yaml
    """
    os.environ["X-GOBLET-DEPLOY"] = "false"
    try:
        if stage:
            os.environ["STAGE"] = stage
        app = get_goblet_app(GConfig().main_file or "main.py")
        permissions = app.get_permissions()
        custom_role_policy = create_custom_role_policy(app.function_name, permissions)
        if yaml:
            with open(".goblet/autogen_iam_role.yaml", "w") as f:
                f.write(yml.dump(custom_role_policy))
        else:
            with open(".goblet/autogen_iam_role.json", "w") as f:
                f.write(json.dumps(custom_role_policy))

    except FileNotFoundError as not_found:
        click.echo(
            f"Missing {not_found.filename}. Make sure you are in the correct directory and this file exists"
        )
        sys.exit(1)


@services.command(name="create_service_account")
@click.option("-n", "--name", "name", envvar="NAME")
@click.option("-p", "--project", "project", envvar="GOOGLE_PROJECT")
@click.option("-s", "--stage", "stage", envvar="STAGE")
def create_service_account(name, project, stage):
    """Create a custom role and service account for deploying the goblet app"""
    os.environ["X-GOBLET-DEPLOY"] = "true"
    try:
        _project = project or get_default_project()
        if not _project:
            click.echo(
                "Project not found. Set --project flag or add to gcloud by using gcloud config set project PROJECT"
            )
        os.environ["GOOGLE_PROJECT"] = _project
        if stage:
            os.environ["STAGE"] = stage
        app = get_goblet_app(GConfig().main_file or "main.py")
        permissions = app.get_permissions()
        custom_role_policy = create_custom_role_policy(
            name or app.function_name, permissions
        )
        app.create_service_account(custom_role_policy)

    except FileNotFoundError as not_found:
        click.echo(
            f"Missing {not_found.filename}. Make sure you are in the correct directory and this file exists"
        )
        sys.exit(1)


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter
